import numpy as np

from .. import errors
from ..segment import Segment
from ..filesystem import VTOC, Dirent, Directory, Filesystem
from ..file_type import guess_file_type
from ..char_mapping import internal_to_atascii, atascii_to_internal
from .atari_dos2 import AtariDos2, AtariDosBootSegment
from ..machines.atari8bit.jumpman import playfield

import logging
log = logging.getLogger(__name__)


class AtariJumpmanBootSegment(AtariDosBootSegment):
    def find_segment_location(self, media):
        self.bldadr = 0x700
        try:
            sectors = list(range(1, 8))  # sectors 1 - 7
            sectors.extend(range(560, 720))  # sectors 560 - 719
            indexes = media.get_sector_list_offsets(sectors)
        except errors.MediaError as e:
            raise errors.IncompatibleMediaError(f"Invalid boot sector: {e}")
        return indexes, 0

    def calc_boot_segments(self):
        header = Segment(self, 0, 0x700, "Boot Header", length=6)
        code = Segment(self, 6, 0x706, name="Boot Code", length=0x380 - 6)
        game = Segment(self, 0x380, 0x2000, name="Game Code", length=0x5000)
        game.segments = [
            Segment(game, 0, 0x2000, name="Game Code, Part 1", length=0x800),
            Segment(game, 0x800, 0x2800, name="Level Definition", length=0x800),
            Segment(game, 0x1000, 0x4000, name="Game Code, Part 2", length=0x4000),
        ]
        return [header, code, game]


class AtariJumpmanDirent(Dirent):
    ui_name = "Jumpman Level"
    name_offset = 0x2bec - 0x2800
    dirent_size = 0x800
    extra_serializable_attributes = ['file_num', 'in_use', 'is_sane', 'id', 'level_name']
    dependent_file_attributes = ['assembly_source']

    def __init__(self, directory, file_num):
        start = file_num * self.dirent_size
        self.id = b''
        self.level_name = b''
        self.assembly_source = None
        Dirent.__init__(self, directory, file_num, start, self.dirent_size)
        self.name = str(self)
        self.origin = 0x2800
        self.restore_computed_defaults()

    def restore_computed_defaults(self):
        self.jumpman_playfield_model = playfield.JumpmanPlayfieldModel(self)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.id == other.id

    def __str__(self):
        return f"Level {self.id.decode('latin1')}: {self.level_name.decode('latin1')}"

    @property
    def filename(self):
        return self.level_name

    def parse_raw_dirent(self):
        data = self.data[:]
        self.id = bytes(data[0:2].view(dtype="S2"))
        name = internal_to_atascii[data[self.name_offset:self.name_offset + 20] & 0x3f]
        self.level_name = bytes(name).strip()

    def encode_dirent(self):
        data = np.zeros([self.dirent_size], dtype=np.uint8)
        values = data.view(dtype=self.format)[0]
        values[0] = self.id
        values[1] = self.level_name
        return data

    def get_file(self):
        return None

    def sanity_check(self):
        return len(self) == 0x800

    def set_assembly_source(self, src):
        """Attach an assembly language source file to this segment, to be
        compiled and used as the custom code for this level.

        Note that it will be automatically recompiled during idle time, and so
        is not immediately compiled when added here.
        """
        self.assembly_source = src

    def update_dependent_file(self, path):
        # Currently, this is assuming that there is only one dependent source
        # file, so as soon as this is triggered, recompile the code.
        log.debug(f"update_dependent_file: recompiling {path}")
        self.jumpman_playfield_model.compile_assembly_source()

    def error_in_dependent_file(self, path, error):
        log.error(f"error_in_dependent_file: {path}: {error}")
        self.jumpman_playfield_model.compile_assembly_source()


class AtariJumpmanDirectory(Directory):
    ui_name = "Jumpman Levels"

    def __str__(self):
        num_entries = len(self) // AtariJumpmanDirent.dirent_size
        s = f"{self.name} ({num_entries} level"
        if num_entries != 1:
            s += "s"
        s += ")"
        if self.error:
            s += " " + self.error
        return s

    def find_first_level(self, media):
        index, size = media.get_contiguous_sectors_offsets(17, 16 * 32)
        return index, size

    def find_segment_location(self):
        media = self.media
        index, size = self.find_first_level(media)
        return index, size

    def calc_dirents(self):
        segments = []
        filenum = 0
        pointer = 0
        while pointer < len(self):
            filenum = pointer // AtariJumpmanDirent.dirent_size
            dirent = AtariJumpmanDirent(self, filenum)
            pointer += AtariJumpmanDirent.dirent_size
            segments.append(dirent)
        return segments


class AtariJumpman(AtariDos2):
    ui_name = "Atari Jumpman"

    def check_media(self, media):
        AtariDos2.check_media(self, media)
        if media.sector_size != 128:
            raise errors.IncompatibleMediaError(f"{self.ui_name} requires SD")

    def calc_boot_segment(self):
        return AtariJumpmanBootSegment(self)

    def calc_vtoc_segment(self):
        return None

    def calc_directory_segment(self):
        return AtariJumpmanDirectory(self)

    def find_interesting_segment(self):
        levels = list(self.iter_dirents())
        if levels:
            return levels[0]
        return None


class AtariJumpmanLevelTesterBootSegment(AtariDosBootSegment):
    def find_segment_location(self, media):
        self.bldadr = 0x700
        start, size = media.get_index_of_sector(4)
        i = 9
        count = media[i] + 256 * media[i+1] + 256 * 256 *media[i + 2]
        if start + count > len(media) or start + count < len(media) - 128:
            raise errors.NotEnoughSpaceOnDisk(f"KBoot header reports size {count}; media only {len(media)}")
        else:
            self.exe_size = count
            self.exe_start = start
        return 0, len(media)

    def calc_boot_segments(self):
        header = Segment(self, 0, 0x700, "Boot Header", length=6)
        code = Segment(self, 6, 0x706, name="Boot Code", length=0x180 - 6)

        file_segment = guess_file_type(self, "Jumpman Level", self.exe_start, self.exe_size)
        return [header, code, file_segment]


class AtariJumpmanLevelTesterDirectory(AtariJumpmanDirectory):
    def find_first_level(self, media):
        start, size = media.get_index_of_sector(4)
        i = 9
        count = media[i] + 256 * media[i+1] + 256 * 256 *media[i + 2]
        if start + count > len(media) or start + count < len(media) - 128:
            raise errors.NotEnoughSpaceOnDisk(f"Jumpman Level Tester header reports size {count}; media only {len(media)}")
        else:
            exe_size = count
            exe_start = start
        file_segment = guess_file_type(media, "Jumpman Level Tester", exe_start, exe_size)
        index = None
        if len(file_segment.segments) > 0:
            s = file_segment.segments[0].segments[0]
            if s.origin == 0x8800:
                # Jumpman level tester loads initial level at $8800, then the
                # code moves it down
                #
                # Need index relative to media; container_offset is absolute
                # position in source file and media is at some offset from the
                # beginning of the container.
                index = s.container_offset[0] - media.container_offset[0]
        if index is None:
            raise errors.FilesystemError("Not recognized as a Jumpman Level Tester image")
        return index, 0x800


class AtariJumpmanLevelTester(AtariJumpman):
    ui_name = "Atari Jumpman Level Tester"

    def calc_boot_segment(self):
        return AtariJumpmanLevelTesterBootSegment(self)

    def calc_directory_segment(self):
        return AtariJumpmanLevelTesterDirectory(self)



import os

import wx

import numpy as np
from atrip import Segment, Container, style_bits
from atrip.machines.atari8bit import powerup_colors

from . import parser as ju

import logging
log = logging.getLogger(__name__)
drawlog = logging.getLogger("refresh")


class JumpmanPlayfieldModel:
    def __init__(self, segment):
        self.segment = segment
        self.possible_jumpman_segment = ju.is_valid_level_segment(self.segment)
        self.items_per_row = 160
        self.antic_lines = 88
        self.playfield = self.get_playfield_segment()
        self.pick_buffer = np.zeros((self.items_per_row * self.antic_lines), dtype=np.int32)
        self.coin_harvest_diff = -1
        self.num_ladders = -1
        self.num_downropes = -1
        self.num_coins = -1
        self.can_select_objects = False
        self.can_erase_objects = False
        self.assembly_error = ""
        self.custom_code = None
        self.old_trigger_mapping = {}
        self.trigger_root = None
        self.current_level = None
        self.cached_screen = None
        self.valid_level = False
        self.force_refresh = False
        self.screen_state = None
        self.trigger_state = None

    @property
    def assembly_source(self):
        return self.segment.assembly_source

    def init_level_builder(self, segment_viewer):
        self.segment_viewer = segment_viewer
        self.segment = segment_viewer.segment
        log.debug(f"init_level_builder: {self.segment}")
        self.possible_jumpman_segment = ju.is_valid_level_segment(self.segment)
        self.level_builder = ju.JumpmanLevelBuilder(self.segment_viewer.document.user_segments)
        self.cached_screen = None
        self.valid_level = False
        self.force_refresh = True
        self.screen_state = None
        self.trigger_state = None
        self.generate_display_objects()
        self.level_colors = self.calc_level_colors()
        self.draw_playfield(True)

    @property
    def mouse_mode(self):
        return self.segment_viewer.control.mouse_mode

    def generate_display_objects(self, resync=False):
        try:
            source, level_addr, harvest_addr = self.get_level_addrs()
            index = level_addr - source.origin
            self.level_builder.parse_level_data(source, level_addr, harvest_addr)
            self.force_refresh = True
            self.valid_level = True
            if self.trigger_root is not None:
                self.trigger_root = self.level_builder.find_equivalent_coin(self.trigger_root)
            if resync:
                self.mouse_mode.resync_objects()
        except RuntimeError:
            self.valid_level = False

    def get_playfield(self):
        data = np.empty(self.items_per_row * self.antic_lines, dtype=np.uint8)
        return data

    def get_playfield_segment(self, playfield=None):
        if playfield is None:
            playfield = self.get_playfield()
        c = Container(playfield)
        return Segment(c, origin=0x7000)

    def clear_playfield(self, playfield=None):
        if playfield is None:
            playfield = self.playfield
        playfield[:] = 8  # background is the 9th ANTIC color register (counting from zero)
        playfield.style[:] = 0

    def calc_level_colors(self):
        if self.valid_level:
            colors = self.segment[0x2a:0x33].copy()
            # on some levels, the bombs are set to color 0 because they are
            # cycled to produce a glowing effect, but that's not visible here
            # so we force it to be bright white
            fg = colors[4:8]
            fg[fg == 0] = 15
        else:
            colors = powerup_colors()
        return list(colors)

    def get_level_addrs(self):
        if not self.possible_jumpman_segment:
            raise RuntimeError
        source = self.segment
        start = source.origin
        level_addr = source[0x37] + source[0x38]*256
        harvest_addr = source[0x4e] + source[0x4f]*256
        log.debug("level def table: %x, harvest table: %x" % (level_addr, harvest_addr))
        last = source.origin + len(source)
        if level_addr > start and harvest_addr > start and level_addr < last and harvest_addr < last:
            return source, level_addr, harvest_addr
        raise RuntimeError

    def set_trigger_root(self, root):
        if root is not None:
            root = self.level_builder.find_equivalent_coin(root)
        log.debug(f"setting trigger root to {root}")
        self.trigger_root = root
        self.force_refresh = True
        self.trigger_state = None

    @property
    def is_editing_trigger(self):
        return self.trigger_root is not None

    def get_screen_state(self):
        if self.trigger_state is not None:
            return self.trigger_state
        return self.screen_state

    def set_current_screen(self, screen=None):
        if screen is None:
            screen = self.playfield
        else:
            self.cached_screen = None
        self.data = screen
        self.style = screen.style

    def redraw_current(self, screen, overlay_objects=[]):
        self.clear_playfield(screen)
        self.pick_buffer[:] = -1
        self.level_builder.set_harvest_offset(self.mouse_mode.get_harvest_offset())
        main_state = self.level_builder.draw_objects(screen, None, self.segment, highlight=overlay_objects, pick_buffer=self.pick_buffer)
        drawlog.debug("draw objects: %s" % self.level_builder.objects)
        drawlog.debug("highlight objects: %s" % overlay_objects)
        if main_state.missing_object_codes:
            log.error("missing draw codes: %s" % (sorted(main_state.missing_object_codes)))
        if self.trigger_root is not None:
            self.level_builder.fade_screen(screen)
            root = [self.trigger_root]
            self.level_builder.draw_objects(screen, root, self.segment, highlight=root, pick_buffer=self.pick_buffer)
            # change highlight to comment color for selected trigger coin so
            # you don't get confused with any objects actually selected
            old_highlight = np.where(screen.style == style_bits.selected_bit_mask)
            screen.style[old_highlight] |= style_bits.comment_bit_mask
            screen.style[:] &= (0xff ^ (style_bits.match_bit_mask|style_bits.selected_bit_mask))

            # replace screen state so that the only pickable objects are
            # those in the triggered layer
            self.pick_buffer[:] = -1
            trigger_state = self.level_builder.draw_objects(screen, self.trigger_root.trigger_painting, self.segment, highlight=overlay_objects, pick_buffer=self.pick_buffer)
            active_state = trigger_state
        else:
            active_state = main_state
            trigger_state = None
        return main_state, trigger_state, active_state

    def compute_image(self, force=False):
        self.set_current_screen()
        if force:
            self.force_refresh = True
        if self.force_refresh or self.cached_screen is None:
            self.screen_state, self.trigger_state, _ = self.redraw_current(self.playfield)
            self.cached_screen = self.playfield[:].copy()
            self.force_refresh = False
            self.update_harvest_state()
        else:
            self.playfield[:] = self.cached_screen
        self.set_current_screen()
        return

    def bad_image(self):
        self.set_current_screen()
        self.playfield[:] = 0
        self.playfield.style[:] = 0
        self.force_refresh = True
        s = self.playfield.container.style.reshape((self.antic_lines, -1))
        s[::2,::2] = style_bits.comment_bit_mask
        s[1::2,1::2] = style_bits.comment_bit_mask
        self.set_current_screen()
        return

    def draw_playfield(self, force=False):
        if self.valid_level:
            override = self.mouse_mode.calc_playfield_override()
            if override is not None:
                drawlog.debug("draw_playfield: using override screen")
                self.set_current_screen(override)
                return
            drawlog.debug("draw_playfield: computing image")
            self.compute_image()
            self.mouse_mode.draw_extra_objects(self.level_builder, self.playfield, self.segment)
        else:
            self.bad_image()
        # self.mouse_mode.draw_overlay(bitimage)  # FIXME!

    ##### Object edit

    def get_save_location(self):
        if self.trigger_root is not None:
            equiv = self.level_builder.find_equivalent_coin(self.trigger_root)
            parent = equiv.trigger_painting
        else:
            parent = None
        return parent

    def delete_objects(self, objects):
        save_location = self.get_save_location()
        self.level_builder.delete_objects(objects, save_location)
        self.segment_viewer.save_changes()

    def save_objects(self, objects):
        save_location = self.get_save_location()
        self.level_builder.add_objects(objects, save_location)
        self.segment_viewer.save_changes(add=True)

    ##### Utilities

    def update_harvest_state(self):
        if not self.current_level.valid_level:
            return
        harvest_state = self.current_level.level_builder.get_harvest_state()
        self.num_ladders = len(harvest_state.ladder_positions)
        self.num_downropes = len(harvest_state.downrope_positions)
        self.num_coins = len(harvest_state.coins)

    def set_current_draw_pattern(self, pattern, control):
        try:
            iter(pattern)
        except TypeError:
            self.draw_pattern = [pattern]
        else:
            self.draw_pattern = pattern

    ##### Custom code handling

    def compile_assembly_source(self):
        log.debug(f"compile_assembly_source: compiling {self.assembly_source}")
        self.custom_code = None
        self.assembly_error = ""
        filename = self.assembly_source
        if not filename:
            return
        self.segment_viewer.editor.metadata_dirty = True
        if not os.path.isabs(filename):
            # relative path; attempt to supply absolute path using document
            d = self.segment_viewer.document
            path = d.filesystem_path()
            if path:
                dirname = os.path.dirname(path)
                filename = os.path.join(dirname, filename)

        try:
            log.debug("compiling jumpman level code in %s" % filename)
            self.custom_code = ju.JumpmanCustomCode(filename)
        except SyntaxError as e:
            self.assembly_error = f"Assembly error:\n{str(e)}"
            log.error(self.assembly_error)
        except ImportError:
            self.assembly_error = f"Assembly error:\nPlease install pyatasm to\ncompile custom code"
            log.error(self.assembly_error)
            self.old_trigger_mapping = dict()
            raise RuntimeError(self.assembly_error)
        if self.custom_code:
            self.update_trigger_mapping()
            self.save_assembly()
        else:
            # error in the code, cause a refresh that will update the code
            # viewer
            self.segment_viewer.editor.document.byte_values_changed_event(flags=True)

    @property
    def custom_code_info(self):
        try:
            if self.assembly_error:
                return self.assembly_error
            return self.custom_code.info
        except AttributeError:
            return "No custom code\n"

    @property
    def action_vector_info(self):
        try:
            if self.assembly_error:
                return ""
            return self.custom_code.vector_summary
        except AttributeError:
            return "No custom code\n"

    @property
    def coin_trigger_info(self):
        try:
            if self.assembly_error:
                return ""
            return self.custom_code.coin_trigger_summary
        except AttributeError:
            return "No custom code\n"

    @property
    def other_label_info(self):
        try:
            if self.assembly_error:
                return ""
            return self.custom_code.label_summary
        except AttributeError:
            return "No custom code\n"

    def save_assembly(self):
        log.debug("save_assembly: code=%s" % self.custom_code)
        code = self.custom_code
        if not code:
            return
        source, level_addr, harvest_addr = self.get_level_addrs()
        ranges, data = code.get_ranges(source)
        self.segment_viewer.save_assembly(source, ranges, data)

    def update_trigger_mapping(self):
        # only create old trigger mapping if one doesn't exist
        if not self.old_trigger_mapping:
            self.old_trigger_mapping = dict(self.get_triggers())
        else:
            old_map = self.old_trigger_mapping
            new_map = self.get_triggers()
            if old_map != new_map:
                log.debug("UPDATING trigger map!")
                log.debug("old map %s" % old_map)
                log.debug("new_map %s" % new_map)
                self.current_level.update_triggers(old_map, new_map)
                # FIXME: what about undo and the trigger mapping?
                self.current_level.save_changes(AssemblyChangedCommand)
                self.old_trigger_mapping = new_map

    def update_harvest_state(self):
        if not self.valid_level:
            return
        harvest_state = self.level_builder.get_harvest_state()
        self.num_ladders = len(harvest_state.ladder_positions)
        self.num_downropes = len(harvest_state.downrope_positions)
        self.num_coins = len(harvest_state.coins)

    ##### Jumpman coin triggers

    def get_triggers(self):
        code = self.custom_code
        if code is None:
            return {}
        return code.triggers

    def get_trigger_label(self, addr):
        rev_old_map = {v: k for k, v in self.old_trigger_mapping.items()}
        if addr in rev_old_map:
            return rev_old_map[addr]
        return None


import time

import numpy as np

from . import libatari800 as liba8
from . import dtypes as d
from . import akey
from .colors import NTSC
from ...emulator import Emulator
from ...errors import FrameNotFinishedError

import logging
log = logging.getLogger(__name__)


debug_frames = False

def clamp(val):
    if val < 0.0:
        return 0
    elif val > 255.0:
        return 255
    return int(val)

ntsc_iq_lookup = [
    [  0.000,  0.000 ],
    [  0.144, -0.189 ],
    [  0.231, -0.081 ],
    [  0.243,  0.032 ],
    [  0.217,  0.121 ],
    [  0.117,  0.216 ],
    [  0.021,  0.233 ],
    [ -0.066,  0.196 ],
    [ -0.139,  0.134 ],
    [ -0.182,  0.062 ],
    [ -0.175, -0.022 ],
    [ -0.136, -0.100 ],
    [ -0.069, -0.150 ],
    [  0.005, -0.159 ],
    [  0.071, -0.125 ],
    [  0.124, -0.089 ],
    ]

def gtia_ntsc_to_rgb_table(val):
    # This is a better representation of the NTSC colors using a lookup table
    # rather than the phase calculations. Also from the same thread:
    # http://atariage.com/forums/topic/107853-need-the-256-colors/page-2#entry1319398
    cr = (val >> 4) & 15;
    lm = val & 15;

    y = 255*(lm+1)/16;
    i = ntsc_iq_lookup[cr][0] * 255
    q = ntsc_iq_lookup[cr][1] * 255

    r = y + 0.956*i + 0.621*q;
    g = y - 0.272*i - 0.647*q;
    b = y - 1.107*i + 1.704*q;

    return clamp(r), clamp(g), clamp(b)

def ntsc_color_map():
    rmap = np.empty(256, dtype=np.uint8)
    gmap = np.empty(256, dtype=np.uint8)
    bmap = np.empty(256, dtype=np.uint8)

    for i in range(256):
        r, g, b = gtia_ntsc_to_rgb_table(i)
        rmap[i] = r
        gmap[i] = g
        bmap[i] = b

    return rmap, gmap, bmap


class Atari800Mixin:
    cpu = "6502"
    name = "atari800"
    ui_name = "Atari 800"

    output_array_dtype = d.OUTPUT_DTYPE
    width = d.VIDEO_WIDTH
    height = d.VIDEO_HEIGHT

    low_level_interface = liba8

    def serialize_to_dict(self):
        if not self.is_frame_finished:
            raise FrameNotFinishedError("atari800 can't save its internal state in the middle of a frame.")
        return Emulator.serialize_to_dict(self)

    def compute_color_map(self):
        self.rmap, self.gmap, self.bmap = ntsc_color_map()

    @property
    def current_cpu_status(self):
        a, p, sp, x, y, _, pc = self.cpu_state

        # Confirm validity by checking raw locations
        cpu_offset = self.output[0]['tag_cpu']
        registers = self.raw_state[cpu_offset:cpu_offset+5]
        assert a == registers[0]
        assert p == registers[1]
        assert sp == registers[2]
        assert x == registers[3]
        assert y == registers[4]
        raw_pc = self.raw_state[self.output[0]['tag_pc']:]
        assert pc == (raw_pc[0] + 256 * raw_pc[1])
        return "A=%02x X=%02x Y=%02x SP=%02x FLAGS=%02x PC=%04x" % (a, x, y, sp, p, pc)

    @property
    def program_counter(self):
        a, p, sp, x, y, _, pc = self.cpu_state
        return pc

    @property
    def current_antic_status(self):
        antic_offset = self.output[0]['tag_antic']
        registers = self.raw_state[antic_offset:antic_offset+d.ANTIC_DTYPE.itemsize]
        return f"antic: {str(registers)}"

    def process_args(self, emu_args):
        # FIXME: need to find a way to turn on/off carts outside of the init
        # routines because boot_from_file restarts the machine with the new
        # file but currently doesn't change cart status.
        if not emu_args:
            emu_args = [
                #"-basic",
                #"-shmem-debug-video",
                #"jumpman.atr"
            ]
        return emu_args

    def generate_save_state_memory_blocks(self):
        s = self.state_start_offset
        for name in self.output[0].dtype.names:
            print(f"offsets: output[{name}] = {self.output[0][name]}")

        self.computed_dtypes = {
            'ANTIC': d.ANTIC_DTYPE,
            'GTIA': d.GTIA_DTYPE,
            'POKEY': d.POKEY_DTYPE,
            'PIA': d.PIA_DTYPE,
        }
        self.segment_starts = {
            'ANTIC': self.output[0]['tag_antic'],
            'GTIA': self.output[0]['tag_gtia'],
            'POKEY': self.output[0]['tag_pia'],
            'PIA': self.output[0]['tag_pokey'],
        }
        segments = [
            (s + self.output[0]['tag_cpu'], 6, 0, "CPU Regisers"),
            (s + self.output[0]['tag_base_ram'], 256*256, 0, "Main Memory"),
            (s + self.output[0]['tag_antic'], d.ANTIC_DTYPE.itemsize, 0, "ANTIC"),
            (s + self.output[0]['tag_gtia'], d.GTIA_DTYPE.itemsize, 0, "GTIA"),
            (s + self.output[0]['tag_pia'], d.PIA_DTYPE.itemsize, 0, "PIA"),
            (s + self.output[0]['tag_pokey'], d.POKEY_DTYPE.itemsize, 0, "POKEY"),
        ]
        self.save_state_memory_blocks.extend(segments)

    def boot_from_file(self, filename):
        log.debug(f"booting {self.ui_name} from {filename}")
        liba8.reboot_with_file(filename)
        self.configure_io_arrays()

    def end_emulation(self):
        pass

    def debug_video(self):
        video_mem = self.output[0]['video'].view(dtype=np.uint8)
        offset = 336*24 + 8
        for y in range(32):
            print("%x:" % offset, end=' ')
            for x in range(8,60):
                c = video_mem[x + offset]
                if (c == 0 or c == '\x00'):
                    print(" ", end=' ')
                elif (c == 0x94 or c == '\x94'):
                    print(".", end=' ')
                elif (c == 0x9a or c == '\x9a'):
                    print("X", end=' ')
                else:
                    try:
                        print(ord(c), end=' ')
                    except TypeError:
                        print(repr(c), end=' ')
            print()
            offset += 336;

    def calc_cpu_data_array(self):
        cpu_offset = self.output[0]['tag_cpu']
        pc_offset = self.output[0]['tag_pc']
        computed_dtype = np.dtype([
            ("A", np.uint8),
            ("P", np.uint8),
            ("SP", np.uint8),
            ("X", np.uint8),
            ("Y", np.uint8),
            ("_", np.uint8, pc_offset - cpu_offset - 5),
            ("PC", '<u2'),
            ])
        raw = self.raw_state[cpu_offset:cpu_offset + computed_dtype.itemsize]
        print(("sizeof raw_array=%d raw=%d dtype=%d" % (len(self.raw_array), len(raw), computed_dtype.itemsize)))
        dataview = raw.view(dtype=computed_dtype)
        return dataview[0]

    def calc_dtype_data(self, segment_name):
        d = np.dtype(self.computed_dtypes[segment_name])
        start = self.segment_starts[segment_name]
        raw = self.raw_state[start:start + d.itemsize]
        return raw.view(dtype=d)[0]

    def calc_main_memory_array(self):
        offset = self.output[0]['tag_base_ram']
        return self.raw_array[offset:offset + 1<<16]

    # Emulator user input functions

    def coldstart(self):
        """Simulate an initial power-on startup.
        """
        super().coldstart()
        self.send_special_key(akey.AKEY_COLDSTART)
        self.configure_io_arrays()

    def warmstart(self):
        """Simulate a warm start; i.e. pressing the system reset button
        """
        self.send_special_key(akey.AKEY_WARMSTART)
        self.configure_io_arrays()

    def keypress(self, ascii_char):
        self.send_char(ord(ascii_char))

    # Utility functions

    def get_color_indexed_screen(self, frame_number=-1):
        if frame_number < 0:
            output = self.output
        else:
            _, output = self.get_history(frame_number)
        raw = output['video'].reshape((self.height, self.width))
        #print "get_raw_screen", frame_number, raw
        return raw

    def get_frame_rgb(self, frame_number=-1):
        raw = self.get_color_indexed_screen(frame_number)
        self.screen_rgb[:,:,0] = self.rmap[raw]
        self.screen_rgb[:,:,1] = self.gmap[raw]
        self.screen_rgb[:,:,2] = self.bmap[raw]
        return self.screen_rgb

    def get_frame_rgba(self, frame_number=-1):
        raw = self.get_color_indexed_screen(frame_number)
        self.screen_rgba[:,:,0] = self.rmap[raw]
        self.screen_rgba[:,:,1] = self.gmap[raw]
        self.screen_rgba[:,:,2] = self.bmap[raw]
        self.screen_rgba[:,:,3] = 255
        return self.screen_rgba

    def get_frame_rgba_opengl(self, frame_number=-1):
        raw = np.flipud(self.get_color_indexed_screen(frame_number))
        self.screen_rgba[:,:,0] = self.rmap[raw]
        self.screen_rgba[:,:,1] = self.gmap[raw]
        self.screen_rgba[:,:,2] = self.bmap[raw]
        self.screen_rgba[:,:,3] = 255
        return self.screen_rgba

    ##### Input routines

    def get_special_key_actions(self):
        return [
            "emu_a8_start_key",
            "emu_a8_select_key",
            "emu_a8_option_key",
        ]

    def send_special_key(self, key_id):
        self.input['keychar'] = 0
        self.input['keycode'] = 0
        self.input['special_key'] = key_id
        if key_id in [2, 3]:
            self.frame_event.append((self.frame_count + 2, self.clear_keys))

    def set_option(self, state):
        self.input['flags'] |= akey.OPTION_FLAG

    def set_select(self, state):
        self.input['flags'] |= akey.SELECT_FLAG

    def set_start(self, state):
        self.input['flags'] |= akey.START_FLAG


class Atari800XLMixin(Atari800Mixin):
    cpu = "6502"
    name = "atari800xl"
    ui_name = "Atari 800XL"

    def process_args(self, emu_args):
        if not emu_args:
            emu_args = [
                "-xl",
            ]
        return emu_args


class Atari5200Mixin(Atari800Mixin):
    cpu = "6502"
    name = "atari5200"
    ui_name = "Atari 5200"

    def process_args(self, emu_args):
        if not emu_args:
            emu_args = [
                "-5200",
            ]
        return emu_args


try:
    import wx

    class wxMixin:
        wx_to_akey = {
            wx.WXK_BACK: akey.AKEY_BACKSPACE,
            wx.WXK_DELETE: akey.AKEY_DELETE_CHAR,
            wx.WXK_INSERT: akey.AKEY_INSERT_CHAR,
            wx.WXK_ESCAPE: akey.AKEY_ESCAPE,
            wx.WXK_END: akey.AKEY_HELP,
            wx.WXK_HOME: akey.AKEY_CLEAR,
            wx.WXK_RETURN: akey.AKEY_RETURN,
            wx.WXK_SPACE: akey.AKEY_SPACE,
            wx.WXK_F7: akey.AKEY_BREAK,
            wx.WXK_PAUSE: akey.AKEY_BREAK,
            96: akey.AKEY_ATARI,  # back tick
            65: akey.AKEY_a,
            66: akey.AKEY_b,
            67: akey.AKEY_c,
            68: akey.AKEY_d,
            69: akey.AKEY_e,
            70: akey.AKEY_f,
            71: akey.AKEY_g,
            72: akey.AKEY_h,
            73: akey.AKEY_i,
            74: akey.AKEY_j,
            75: akey.AKEY_k,
            76: akey.AKEY_l,
            77: akey.AKEY_m,
            78: akey.AKEY_n,
            79: akey.AKEY_o,
            80: akey.AKEY_p,
            81: akey.AKEY_q,
            82: akey.AKEY_r,
            83: akey.AKEY_s,
            84: akey.AKEY_t,
            85: akey.AKEY_u,
            86: akey.AKEY_v,
            87: akey.AKEY_w,
            88: akey.AKEY_x,
            89: akey.AKEY_y,
            90: akey.AKEY_z,
            48: akey.AKEY_0,
            49: akey.AKEY_1,
            50: akey.AKEY_2,
            51: akey.AKEY_3,
            52: akey.AKEY_4,
            53: akey.AKEY_5,
            54: akey.AKEY_6,
            55: akey.AKEY_7,
            56: akey.AKEY_8,
            57: akey.AKEY_9,
        }

        wx_to_akey_shift = {
            65: akey.AKEY_A,
            66: akey.AKEY_B,
            67: akey.AKEY_C,
            68: akey.AKEY_D,
            69: akey.AKEY_E,
            70: akey.AKEY_F,
            71: akey.AKEY_G,
            72: akey.AKEY_H,
            73: akey.AKEY_I,
            74: akey.AKEY_J,
            75: akey.AKEY_K,
            76: akey.AKEY_L,
            77: akey.AKEY_M,
            78: akey.AKEY_N,
            79: akey.AKEY_O,
            80: akey.AKEY_P,
            81: akey.AKEY_Q,
            82: akey.AKEY_R,
            83: akey.AKEY_S,
            84: akey.AKEY_T,
            85: akey.AKEY_U,
            86: akey.AKEY_V,
            87: akey.AKEY_W,
            88: akey.AKEY_X,
            89: akey.AKEY_Y,
            90: akey.AKEY_Z,
        }

        wx_to_akey_ctrl = {
            wx.WXK_UP: akey.AKEY_UP,
            wx.WXK_DOWN: akey.AKEY_DOWN,
            wx.WXK_LEFT: akey.AKEY_LEFT,
            wx.WXK_RIGHT: akey.AKEY_RIGHT,
            65: akey.AKEY_CTRL_a,
            66: akey.AKEY_CTRL_b,
            67: akey.AKEY_CTRL_c,
            68: akey.AKEY_CTRL_d,
            69: akey.AKEY_CTRL_e,
            70: akey.AKEY_CTRL_f,
            71: akey.AKEY_CTRL_g,
            72: akey.AKEY_CTRL_h,
            73: akey.AKEY_CTRL_i,
            74: akey.AKEY_CTRL_j,
            75: akey.AKEY_CTRL_k,
            76: akey.AKEY_CTRL_l,
            77: akey.AKEY_CTRL_m,
            78: akey.AKEY_CTRL_n,
            79: akey.AKEY_CTRL_o,
            80: akey.AKEY_CTRL_p,
            81: akey.AKEY_CTRL_q,
            82: akey.AKEY_CTRL_r,
            83: akey.AKEY_CTRL_s,
            84: akey.AKEY_CTRL_t,
            85: akey.AKEY_CTRL_u,
            86: akey.AKEY_CTRL_v,
            87: akey.AKEY_CTRL_w,
            88: akey.AKEY_CTRL_x,
            89: akey.AKEY_CTRL_y,
            90: akey.AKEY_CTRL_z,
        }

        def process_key_down(self, evt, keycode):
            log.debug("key down! key=%s mod=%s" % (evt.GetKeyCode(), evt.GetModifiers()))
            key = evt.GetKeyCode()
            mod = evt.GetModifiers()
            if mod == wx.MOD_CONTROL:
                akey = self.wx_to_akey_ctrl.get(key, None)
            elif mod == wx.MOD_SHIFT:
                akey = self.wx_to_akey_shift.get(key, None)
            else:
                akey = self.wx_to_akey.get(key, None)

            if akey is None:
                evt.Skip()
            else:
                self.send_keycode(akey)

        def process_key_up(self, evt, keycode):
            log.debug("key up! key=%s mod=%s" % (evt.GetKeyCode(), evt.GetModifiers()))
            key = evt.GetKeyCode()
            mod = evt.GetModifiers()
            if mod == wx.MOD_CONTROL:
                akey = self.wx_to_akey_ctrl.get(key, None)
            elif mod == wx.MOD_SHIFT:
                akey = self.wx_to_akey_shift.get(key, None)
            else:
                akey = self.wx_to_akey.get(key, None)

            if akey is None:
                evt.Skip()
            else:
                self.send_keycode(0)

        def process_key_state(self):
            try:
                up = 0b0001 if wx.GetKeyState(wx.WXK_UP) else 0
            except wx._core.PyNoAppError:
                return
            down = 0b0010 if wx.GetKeyState(wx.WXK_DOWN) else 0
            left = 0b0100 if wx.GetKeyState(wx.WXK_LEFT) else 0
            right = 0b1000 if wx.GetKeyState(wx.WXK_RIGHT) else 0
            self.input['joysticks'][0] = up | down | left | right

            trig = 0
            if wx.GetKeyState(wx.WXK_TAB):
                trig |= 1
            self.input['joystick_triggers'] = trig
            # print("joy", self.input['joy0'], "trig", self.input['trig0'])

            # console keys will reflect being pressed if at any time between
            # frames the key has been pressed
            mod = 0
            if wx.GetKeyState(wx.WXK_F2) or self.forced_modifier=='option':
                mod |= akey.OPTION_FLAG
            if wx.GetKeyState(wx.WXK_F3) or self.forced_modifier=='select':
                mod |= akey.SELECT_FLAG
            if wx.GetKeyState(wx.WXK_F4) or self.forced_modifier=='start':
                mod |= akey.START_FLAG
            self.input['flags'] = mod

except ImportError:
    class wxMixin:
        pass


class Atari800(wxMixin, Atari800Mixin, Emulator):
    supported_filesystems = ["Atari DOS2", "Atari KBoot"]
    supported_binaries = ["Atari 8-bit Object File"]

class Atari800XL(wxMixin, Atari800XLMixin, Emulator):
    supported_filesystems = ["Atari DOS2", "Atari KBoot", "Atari Jumpman", "Atari Jumpman Level Tester"]
    supported_binaries = ["Atari 8-bit Object File"]

class Atari5200(wxMixin, Atari5200Mixin, Emulator):
    pass



import numpy as np

from sawx.utils.command import Command, UndoInfo
from sawx.utils.sortutil import ranges_to_indexes, indexes_to_ranges
from sawx.utils.permute import bit_reverse_table

import logging
log = logging.getLogger(__name__)
progress_log = logging.getLogger("progress")


class SegmentCommand(Command):
    short_name = "segment_data_base"
    ui_name = "Segment Modification Abstract Command"
    serialize_order =  [
            ('segment', 'int'),
            ]

    def __init__(self, segment):
        Command.__init__(self)
        self.segment = segment

    def coalesce(self, next_command):
        if next_command.__class__ == self.__class__ and next_command.segment == self.segment:
            if self.can_coalesce(next_command):
                self.coalesce_merge(next_command)


class ChangeByteValuesCommand(SegmentCommand):
    short_name = "change_values_base"
    ui_name = "Change Values Abstract Command"

    def __init__(self, segment, advance=False):
        SegmentCommand.__init__(self, segment)
        self.advance = advance

    def set_undo_flags(self, flags):
        flags.byte_values_changed = True


class ChangeMetadataCommand(SegmentCommand):
    short_name = "metadata_base"
    ui_name = "Change Metadata Abstract Command"

    def set_undo_flags(self, flags):
        flags.byte_style_changed = True


class SetSelectionCommand(ChangeByteValuesCommand):
    short_name = "set_indexes_value"
    ui_name = "Set Values at Indexes"
    serialize_order =  [
            ('segment', 'int'),
            ('selection', 'selection'),
            ('data', 'string'),
            ]

    def __init__(self, segment, selection, data=0, advance=False):
        ChangeByteValuesCommand.__init__(self, segment, advance)
        self.selection = selection.copy()  # use FrozenSelection so it won't change after undo/redo
        self.data = data

    def __str__(self):
        return f"{self.ui_name}"

    def get_indexes(self):
        return self.selection.calc_indexes(self.segment, True)

    def get_data(self, orig):
        return self.data

    def change_data_at_indexes(self, indexes):
        old_data = self.segment[indexes].copy()
        self.segment[indexes] = self.get_data(old_data)
        return (old_data, indexes)

    def do_change(self, editor, undo):
        indexes = self.get_indexes()
        i1 = min(indexes)
        i2 = max(indexes)
        log.debug(f"{self.short_name}: do_change: indexes={indexes}")
        undo.flags.index_range = i1, i2
        if self.advance:
            undo.flags.advance_caret_position_in_control = editor.focused_viewer.control
        return self.change_data_at_indexes(indexes)

    def undo_change(self, editor, old_data):
        old_data, indexes = old_data
        self.segment[indexes] = old_data
















class SetContiguousDataCommand(ChangeByteValuesCommand):
    short_name = "set_data_base"
    ui_name = "Set Data Abstract Command"
    serialize_order =  [
            ('segment', 'int'),
            ('start_index', 'int'),
            ('end_index', 'int'),
            ]

    def __init__(self, segment, start_index, end_index, advance=False):
        ChangeByteValuesCommand.__init__(self, segment, advance)
        self.start_index = start_index
        if start_index == end_index:
            end_index += 1
        self.end_index = end_index
        self.caret_at_end = False
        self.ignore_if_same_bytes = False

    def __str__(self):
        if self.end_index - self.start_index > 1:
            return "%s @ %04x-%04x" % (self.ui_name, self.start_index + self.segment.origin, self.end_index + self.segment.origin)
        else:
            return "%s @ %04x" % (self.ui_name, self.start_index)

    def get_data(self, orig):
        raise NotImplementedError

    def do_change(self, editor, undo):
        i1 = self.start_index
        i2 = self.end_index
        undo.flags.index_range = i1, i2
        if self.caret_at_end:
            undo.flags.caret_index = i2
        if self.advance:
            undo.flags.advance_caret_position_in_control = editor.focused_viewer.control
        old_data = self.segment[i1:i2].copy()
        self.segment[i1:i2] = self.get_data(old_data)
        if self.ignore_if_same_bytes and self.segment[i1:i2] == old_data:
            undo.flags.success = False
        return (old_data, )

    def undo_change(self, editor, old_data):
        old_data, = old_data
        self.segment[self.start_index:self.end_index] = old_data


class ChangeByteCommand(SetContiguousDataCommand):
    short_name = "cb"
    ui_name = "Change Bytes"
    serialize_order =  [
            ('segment', 'int'),
            ('start_index', 'int'),
            ('end_index', 'int'),
            ('byte_values', 'string'),
            ('caret_at_end', 'bool'),
            ('ignore_if_same_bytes', 'bool'),
            ]

    def __init__(self, segment, start_index, end_index, byte_values, caret_at_end=False, ignore_if_same_bytes=False, advance=False):
        SetContiguousDataCommand.__init__(self, segment, start_index, end_index, advance=advance)
        self.data = byte_values
        self.caret_at_end = caret_at_end
        self.ignore_if_same_bytes = ignore_if_same_bytes

    def get_data(self, orig):
        return self.data


class CoalescingChangeByteCommand(ChangeByteCommand):
    short_name = "ccb"

    def can_coalesce(self, next_command):
        return next_command.start_index == self.start_index and next_command.end_index == self.end_index

    def coalesce_merge(self, next_command):
        self.data = next_command.data


class SetRangeCommand(ChangeByteValuesCommand):
    short_name = "set_range_base"
    ui_name = "Set Ranges Abstract Command"
    serialize_order =  [
            ('segment', 'int'),
            ('ranges', 'int_list'),
            ('range_to_index_function', 'func_pointer'),
            ]

    def __init__(self, segment, ranges, advance=False, range_to_index_function=None):
        ChangeByteValuesCommand.__init__(self, segment, advance)
        self.ranges = tuple(ranges)

        # function to convert ranges to indexes will be set the first time this
        # command is performed so it can use the function appropriate to the
        # currently focused viewer.
        self.range_to_index_function = range_to_index_function

    def get_data(self, orig):
        raise NotImplementedError

    def perform(self, editor, undo_info):
        if self.range_to_index_function is None:
            self.range_to_index_function = editor.focused_viewer.range_processor
        ChangeByteValuesCommand.perform(self, editor, undo_info)

    def do_change(self, editor, undo):
        indexes = self.range_to_index_function(self.ranges)
        # print(f"{self.short_name}: ranges={self.ranges}, indexes={indexes}")
        undo.flags.index_range = indexes[0], indexes[-1]
        old_data = self.segment[indexes].copy()
        self.segment[indexes] = self.get_data(old_data)
        if self.advance:
            undo.flags.advance_caret_position_in_control = editor.focused_viewer.control
        return old_data

    def undo_change(self, editor, old_data):
        indexes = self.range_to_index_function(self.ranges)
        self.segment[indexes] = old_data


class SetRangeValueCommand(SetRangeCommand):
    short_name = "set_range_value"
    ui_name = "Set Ranges To Value"
    serialize_order =  [
            ('segment', 'int'),
            ('ranges', 'int_list'),
            ('data', 'string'),
            ]

    def __init__(self, segment, ranges, data, advance=False, range_to_index_function=None):
        SetRangeCommand.__init__(self, segment, ranges, advance, range_to_index_function)
        self.data = data

    def get_data(self, orig):
        return self.data


class SetRangeValueModifyIndexesCommand(SetRangeCommand):
    short_name = "set_range_value_modify_index"
    ui_name = "Set Ranges To Value + Modify"
    serialize_order =  [
            ('segment', 'int'),
            ('ranges', 'int_list'),
            ('data', 'string'),
            ]

    def __init__(self, segment, ranges, data, advance=False, range_to_index_function=None):
        SetRangeCommand.__init__(self, segment, ranges, advance, range_to_index_function)
        self.data = data

    def get_data_and_indexes(self, indexes):
        return self.data, indexes

    def do_change(self, editor, undo):
        indexes = self.range_to_index_function(self.ranges)
        print(f"{self.short_name}: ranges={self.ranges}, indexes={indexes}")
        undo.flags.index_range = indexes[0], indexes[-1]
        data, new_indexes = self.get_data_and_indexes(indexes)
        old_data = self.segment[new_indexes].copy()
        self.segment[new_indexes] = data
        if self.advance:
            undo.flags.advance_caret_position_in_control = editor.focused_viewer.control
        return old_data


class ChangeStyleCommand(SetContiguousDataCommand):
    short_name = "cs"
    ui_name = "Change Style"

    def __init__(self, segment):
        start_index = 0
        end_index = len(segment)
        SetContiguousDataCommand.__init__(self, segment, start_index, end_index)

    def set_undo_flags(self, flags):
        flags.byte_style_changed = True

    def get_style(self, editor):
        style = self.segment.style[self.start_index:self.end_index].copy()
        return style

    def clip(self, style):
        if len(style) < self.end_index - self.start_index:
            self.end_index = self.start_index + len(style)

    def update_is_tracing(self, viewer):
        pass

    def do_change(self, editor, undo):
        viewer = editor.focused_viewer
        old_is_tracing = viewer.is_tracing
        new_style = self.get_style(editor)
        self.clip(new_style)
        old_style = self.segment.style[self.start_index:self.end_index].copy()
        self.segment.style[self.start_index:self.end_index] = new_style
        self.update_is_tracing(viewer)
        editor.document.change_count += 1
        return (old_style, old_is_tracing)

    def undo_change(self, editor, old_data):
        old_style, old_is_tracing = old_data
        self.segment.style[self.start_index:self.end_index] = old_style
        viewer.is_tracing = old_is_tracing



class OmnivoreError(RuntimeError):
    """Base class for all errors custom to Omnivore
    """
    pass


class EmulatorError(OmnivoreError):
    """Base class for all emulation errors
    """
    pass


class UnknownEmulatorError(EmulatorError):
    """Raised when the requested emulator is not available
    """
    pass


class EmulatorInUseError(EmulatorError):
    """Raised when the requested emulator already has an active instance. Most
    emulators' low-level code doesn't allow for multiple concurrently-running
    emulation processes.
    """
    pass


class FrameNotFinishedError(EmulatorError):
    """Raised when an operation that must occur between frames is
    attempted while in the middle of a frame.
    """
    pass


class UnknownAssemblerError(OmnivoreError):
    """Raised when the requested assembler is not available
    """
    pass



import numpy as np

from sawx.persistence import get_template


class JumpmanSaveAsATR(object):
    ##### Segment saver interface for menu item display

    export_data_name = "Jumpman Level Tester ATR"
    export_extensions = [".atr"]

    @classmethod
    def encode_data(cls, segment, editor):
        """Segment saver interface: take a segment and produce a byte
        representation to save to disk.
        """
        image = get_template("jumpman_level_tester.atr")
        if image is None:
            raise RuntimeError("Can't find Jumpman Level template file")
        raw = np.fromstring(image, dtype=np.uint8)
        raw[0x0196:0x0996] = segment[:]
        return raw.tobytes()


class JumpmanSaveAsXEX(object):
    ##### Segment saver interface for menu item display

    export_data_name = "Jumpman Level Tester XEX"
    export_extensions = [".xex"]

    @classmethod
    def encode_data(cls, segment, editor):
        """Segment saver interface: take a segment and produce a byte
        representation to save to disk.
        """
        image = get_template("jumpman_level_tester.atr")
        if image is None:
            raise RuntimeError("Can't find Jumpman Level template file")
        raw = np.fromstring(image, dtype=np.uint8)
        raw[0x0196:0x0996] = segment[:]

        # the level tester atr images is a KBoot image, so the XEX is embedded
        # inside it and we can just chop off the initial header
        xex = raw[0x0190:]
        return xex.tobytes()



import wx

import numpy as np

from sawx.ui import compactgrid as cg
from sawx.keybindings import KeyBindingControlMixin
from sawx.ui.mouse_mode import MouseMode
from sawx.utils.command import DisplayFlags
from sawx.utils.sortutil import ranges_to_indexes, collapse_overlapping_ranges
from .selection import CurrentSelection
# from ..arch.disasm import get_style_name
# from sawx.framework import actions as fa
# from ..byte_edit import actions as ba
# from ..viewers import actions as va
from ..commands import SetSelectionCommand
from .. import clipboard_helpers

import logging
log = logging.getLogger(__name__)


class SegmentTable(cg.HexTable):
    def __init__(self, linked_base, items_per_row, row_labels_in_multiples=True):
        self.linked_base = linked_base
        cg.HexTable.__init__(self, None, None, items_per_row, self.segment.origin, row_labels_in_multiples)

    @property
    def segment(self):
        return self.linked_base.segment

    @property
    def document(self):
        return self.linked_base.document

    @property
    def data(self):
        return self.linked_base.segment.data

    @property
    def style(self):
        return self.linked_base.segment.style

    def get_label_at_index(self, index):
        # Can't just return hex value of index because some segments (like the
        # raw sector segment) use different labels
        return self.segment.get_ui_name_at_index(index, True)


class SegmentVirtualTable(cg.HexTable):
    col_labels = []
    col_sizes = []

    def __init__(self, linked_base, bytes_per_row_ignored=16):
        self.linked_base = linked_base
        data, style = self.get_data_style_view(linked_base)
        num_cols = self.calc_num_cols()
        cg.HexTable.__init__(self, data, style, num_cols)
        if style is None:
            self.style = np.zeros(self.last_valid_index, dtype=np.uint8)

    @property
    def segment(self):
        return self.linked_base.segment

    def get_data_style_view(self, linked_base):
        """Subclasses implement this to return arrays to be used for data and
        style.

        If style is returned as None, a style array will be created based on
        the largest index of the data.
        """
        raise NotImplementedError

    def calc_num_cols(self):
        return len(self.col_labels)

    def get_value_style(self, row, col):
        raise NotImplementedError

    def get_label_at_index(self, index):
        raise NotImplementedError


class SegmentGridControl(KeyBindingControlMixin, cg.CompactGrid):
    default_table_cls = SegmentTable

    keybinding_desc = {
        "caret_move_up": "Up",
        "caret_move_down": "Down",
        "caret_move_left": "Left",
        "caret_move_right": "Right",
        "caret_move_page_down": "Pagedown",
        "caret_move_page_up": "Pageup",
        "caret_move_start_of_line": "Home",
        "caret_move_end_of_line": "End",
        "caret_move_next_line": "Tab",
        "advance_caret_position": "Space",
    }

    def __init__(self, parent, linked_base, mdict, viewer_cls):
        self.original_metadata = mdict.copy()

        if viewer_cls.override_table_cls is not None:
            # instance attribute will override class attribute
            self.default_table_cls = viewer_cls.override_table_cls

        self.view_params = linked_base.editor.preferences
        self.set_view_param_defaults()
        table = self.calc_default_table(linked_base)

        cg.CompactGrid.__init__(self, table, linked_base.editor.preferences, None, viewer_cls.default_mouse_mode_cls, parent)
        KeyBindingControlMixin.__init__(self)
        # self.automatic_refresh = False

    def map_char_events(self):
        # force keys to be bound on main grid, not parent (which is a panel)
        KeyBindingControlMixin.map_char_events(self, self.main)

    # def map_mouse_events(self):
    #     MouseEventMixin.map_mouse_events(self, self.main)
    #     self.Bind(wx.EVT_MOUSEWHEEL, self.on_mouse_wheel)

    def set_view_param_defaults(self):
        # called from base class to set up initial viewer
        cg.CompactGrid.set_view_param_defaults(self)
        self.set_viewer_defaults()
        self.restore_extra_from_dict(self.original_metadata)

    def set_viewer_defaults(self):
        self.items_per_row = 99  # unlikely to be used normally, so something is likely wrong if a viewer ever appears with 99 colums
        self.zoom = 1

    def serialize_extra_to_dict(self, mdict):
        """Save extra metadata to a dict so that it can be serialized
        """
        mdict['items_per_row'] = self.items_per_row
        mdict['zoom'] = self.zoom

    def restore_extra_from_dict(self, e):
        log.debug("metadata: %s" % str(e))
        if 'items_per_row' in e:
            value = e['items_per_row']
            if value > 0:
                self.items_per_row = value
        if 'zoom' in e:
            value = e['zoom']
            if value > 0:
                self.zoom = value

    def calc_default_table(self, linked_base):
        return self.default_table_cls(linked_base, self.items_per_row)

    def calc_caret_handler(self):
        return CurrentSelection(self)

    def recalc_view(self):
        # just recreate everything. If a subclass wants something different,
        # let it do it itself.
        try:
            self.view_params = self.segment_viewer.linked_base.editor.preferences
        except AttributeError:
            log.warning("segment_viewer not set in recalc_view (probably not a real problem; likely a trait event before setup has been completed)")
        else:
            self.table = self.calc_default_table(self.segment_viewer.linked_base)
            self.recalc_line_renderer()

    def recalc_line_renderer(self):
        self.line_renderer = self.calc_line_renderer()
        log.debug("recalculating %s; items_per_row=%s" % (self, self.items_per_row))
        cg.CompactGrid.recalc_view(self)

    def verify_line_renderer(self):
        # hook for subclass to recreate the line renderer if it needs data from
        # the Machine (which isn't created until after the control and segment
        # viewer instances are created)
        pass

    @property
    def page_size(self):
        """Number of rows in a page, used for scroll-by-page functions"""
        return self.main.fully_visible_rows - 1

    ##### Caret handling

    def stop_scroll_timer(self):
        self.main.scroll_timer.Stop()

    # FIXME: temporary hack until compactgrid uses indexes directly
    def caret_indexes_to_display_coords(self):
        row, col = self.table.index_to_row_col(self.caret_handler.caret_index)
        self.main.show_caret(col, row)

    #### event handling

    def create_char_event_flags(self):
        flags = DisplayFlags(self)
        flags.old_carets = set(self.caret_handler.calc_state())
        return flags

    def do_char_action(self, evt, action):
        flags = self.create_char_event_flags()
        flags.source_control = self
        action(evt, flags)
        self.keep_current_caret_on_screen(flags)
        self.commit_change(flags)

    def do_char_ordinary(self, evt):
        c = evt.GetKeyCode()
        print(f"do_char_ordinary: {c} for {self}")
        if not self.is_editing_in_cell:
            if self.verify_keycode_can_start_edit(c):
                print(f"do_char_ordinary: {c} starts editing!")
                self.start_editing(evt)
            else:
                print(f"do_char_ordinary: {c} not valid to start editing in cell")
                evt.Skip()
        else:
            print(f"do_char_ordinary: editing in cell")
            self.edit_source.EmulateKeyPress(evt)

    def commit_change(self, flags):
        flags.advance_caret_position_in_control = self
        flags.sync_caret_from_control = self

        log.debug(f"\n\ncommit before: {self.caret_handler.carets} flags={flags}")
        linked_base = self.segment_viewer.linked_base
        self.segment_viewer.update_current_selection(self.caret_handler)
        self.mouse_mode.refresh_ranges(self.caret_handler)
        if flags.viewport_origin is not None:
            self.move_viewport_origin(flags.viewport_origin)
        linked_base.sync_caret_event(flags=flags)
        linked_base.ensure_visible_event(flags=flags)
        linked_base.document.refresh_event(flags=flags)
        log.debug(f"commit after: {self.caret_handler.carets} flags={flags}\n\n")

    def process_flags(self, flags):
        self.caret_handler.process_char_flags(flags)

    def post_process_caret_flags(self, flags):
        """Perform any caret updates after the data model has been regenerated
        (e.g. the disassembler where the number of bytes per row can change
        after an edit)

        """
        log.debug(f"post_process_caret_flags: advancing to next position {str(flags)}")
        ch = self.caret_handler
        selection_before = ch.has_selection
        self.advance_caret_position(None, flags)
        ch.validate_carets()
        if selection_before:
            ch.collapse_selections_to_carets()
            ch.refresh_style_from_selection(self.table)
        flags.sync_caret_from_control = self


    ##### Rectangular regions

    def get_rects_from_selections(self):
        rects = []
        for caret in self.caret_handler.carets_with_selection:
            if caret.rectangular:
                rects.append(caret.range)
        return rects

    def get_data_from_rect(self, r):
        (r1, c1), (r2, c2) = r
        last = r2 * self.table.items_per_row
        d = self.segment_viewer.segment[:last].reshape(-1, self.table.items_per_row)
        data = d[r1:r2, c1:c2]
        return r2 - r1, c2 - c1, data

    ##### Copy/paste

    def calc_clipboard_data_objs(self, focused_control):
        segment = self.table.segment
        ranges = self.get_selected_ranges()
        indexes = ranges_to_indexes(ranges)
        log.debug(f"calc_clipboard_data_objs: {self}, ranges={ranges} indexes={indexes}")
        data_objs = []
        if len(ranges) > 0:
            blob = clipboard_helpers.create_numpy_clipboard_blob(ranges, indexes, segment, self)
            data_objs.append(blob.data_obj)
            data_objs.append(blob.text_data_obj(self.view_params.text_copy_stringifier))
            # elif np.alen(indexes) > 0:
            #     data = segment[indexes]
            #     s1 = data.tobytes()
            #     s2 = indexes.tobytes()
            #     serialized = b"%d,%d,%s%s%s" % (len(s1), len(s2), s1, s2, metadata)
            #     name = "numpy,multiple"

        return data_objs


        # rects = self.get_rects_from_selections()
        # data = []
        # for (r1, c1), (r2, c2) in rects:
        #     start, _ = self.table.get_index_range(r1, c1)
        #     _, end = self.table.get_index_range(r2, c2)
        #     data.append((start, end, self.table.data[start:end]))
        # if len(data) > 1:
        #     data_obj = wx.CustomDataObject("numpy,multiple")
        #     # multiple





    def get_start_end_index_of_row(self, row):
        return self.table.get_start_end_index_of_row(row)

    def get_status_at_index(self, index):
        if self.table.is_index_valid(index):
            label = self.table.get_label_at_index(index)
            message = self.get_status_message_at_index(index)
            return "%s: %s %s" % (self.segment_viewer.name, label, message)
        return ""

    def get_status_message_at_index(self, index):
        if self.table.is_index_valid(index):
            s = self.segment_viewer.linked_base.segment
            msg = get_style_name(s, index)
            comments = s.get_comment_at(index)
            return "%s  %s" % (msg, comments)
        return ""

    def get_status_message_at_cell(self, row, col):
        try:
            r, c, index, index2 = self.main.enforce_valid_caret(row, col)
            return self.get_status_at_index(index)
        except IndexError:
            pass
        return ""

    def show_popup(self, popup_desc, data):
        self.segment_viewer.editor.show_popup(popup_desc, data)

    def add_popup_data(self, evt, popup_data):
        popup_data["popup_viewer"] = self.segment_viewer

    def calc_popup_menu(self, evt):
        return self.segment_viewer.popup_menu_desc

    ##### editing

    def verify_keycode_can_start_edit(self, c):
        return False

    def mouse_event_in_edit_cell(self, evt):
        r, c, _ = self.get_row_col_from_event(evt)
        index, _ = self.table.get_index_range(r, c)
        print(f"mouse_event_in_edit_cell: {r},{c}, index={index}")
        return self.caret_handler.is_index_of_caret(index)

    def on_left_down_in_edit_cell(self, evt):
        pass

    def on_motion_in_edit_cell(self, evt):
        pass

    def on_left_up_in_edit_cell(self, evt):
        pass

    def start_editing(self, evt):
        self.is_editing_in_cell = True
        self.edit_source = self.create_hidden_text_ctrl()
        self.edit_source.SetFocus()
        if self.use_first_char_when_starting_edit():
            print(f"start_editing: EmulateKeyPress: {evt.GetKeyCode()} for {self.edit_source}")
            self.edit_source.EmulateKeyPress(evt)

    def use_first_char_when_starting_edit(self):
        return True

    def accept_edit(self, autoadvance=False):
        val = self.edit_source.get_processed_value()
        self.end_editing()
        print(f"accept_edit: changing to {val}")
        self.process_edit(val)

    def process_edit(self, val):
        cmd = self.calc_edit_command(val)
        flags = self.segment_viewer.editor.process_command(cmd)

    def calc_edit_command(self, val):
        cmd = SetSelectionCommand(self.segment_viewer.segment, self.caret_handler, val, advance=True)
        return cmd

    def end_editing(self):
        if self.is_editing_in_cell:
            self.edit_source.Destroy()
            self.edit_source = None
            self.is_editing_in_cell = False
            self.SetFocus()

    def create_hidden_text_ctrl(self):
        c = cg.GridCellTextCtrl(self, -1, 0, pos=(600,100), size=(400,24))
        return c


class SegmentVirtualGridControl(SegmentGridControl):
    default_table_cls = SegmentVirtualTable

    def calc_default_table(self, linked_base):
        table = self.default_table_cls(linked_base)
        self.items_per_row = table.items_per_row
        self.want_row_header = table.want_row_header
        self.want_col_header = table.want_col_header
        return table

    def calc_line_renderer(self):
        return cg.VirtualTableLineRenderer(self, 1, widths=self.table.col_sizes, col_labels=self.table.col_labels)

    ##### editing

    def set_viewer_defaults(self):
        self.items_per_row = -1



import sys
import wx

# from .actions import *

import logging
log = logging.getLogger(__name__)


class SegmentList(wx.ListBox):
    """Segment selector for choosing which portion of the binary data to view
    """

    def __init__(self, parent, linked_base, mdict, viewer_cls, **kwargs):
        self.linked_base = linked_base

        wx.ListBox.__init__(self, parent, style=wx.LB_SINGLE, **kwargs)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        self.Bind(wx.EVT_LISTBOX, self.on_click)
        self.Bind(wx.EVT_LISTBOX_DCLICK, self.on_dclick)
        self.Bind(wx.EVT_RIGHT_DOWN, self.on_popup)
        self.Bind(wx.EVT_MOTION, self.on_tooltip)
        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.Bind(wx.EVT_KEY_UP, self.on_key_up)
        if sys.platform.startswith("linux") and not parent.IsTopLevel():
            # hack needed to force the main window to get Ctrl-F keystrokes. If
            # this ListBox is a direct child of a top-level frame, this is not
            # needed as it won't be visible all the time and therefore won't
            # interfere with the normal Ctrl-F find processing
            self.ui_action = wx.UIActionSimulator()
            self.Bind(wx.EVT_CHAR_HOOK, self.on_char_hook_find_hack)
        self.index_to_segment_uuid = []
        self.segment_uuid_to_index = {}

    def DoGetBestSize(self):
        """ Base class virtual method for sizer use to get the best size
        """
        width = 300
        height = -1
        best = wx.Size(width, height)

        # Cache the best size so it doesn't need to be calculated again,
        # at least until some properties of the window change
        self.CacheBestSize(best)

        return best

    def recalc_view(self):
        c = self.linked_base.document.collection
        self.set_segments(c, self.linked_base.segment_uuid)

    def refresh_view(self):
        self.Refresh()

    def show_segment_in_list(self, segment):
        return hasattr(segment, "container")

    def filter_segments(self, collection, selected=None):
        self.index_to_segment_uuid = []
        self.segment_uuid_to_index = {}
        names = []
        found = 0
        for segment_index, (segment, level) in enumerate(collection.iter_menu()):
            if self.show_segment_in_list(segment):
                if selected == segment.uuid:
                    found = len(self.index_to_segment_uuid)
                self.index_to_segment_uuid.append(segment.uuid)
                self.segment_uuid_to_index[segment.uuid] = len(names)
                names.append((level - 1) * "    " + str(segment))
        return names, found

    def set_segments(self, collection, selected=0):
        items, selected = self.filter_segments(collection, selected)
        if len(items) != self.GetCount():
            self.SetItems(items)
        else:
            for i, item in enumerate(items):
                old = self.GetString(i)
                if old != item:
                    self.SetString(i, item)
        self.SetSelection(selected)

    def on_char_hook_find_hack(self, evt):
        log.debug("on_char_hook_find_hack! char=%s, key=%s, modifiers=%s" % (evt.GetUnicodeKey(), evt.GetKeyCode(), bin(evt.GetModifiers())))
        if evt.GetUnicodeKey() == 70 and evt.ControlDown():
            log.debug("on_char_hook_find_hack! redirecting to %s" % self.linked_base.editor.control)
            # On linux/GTK, the ctrl-F is bound to the list's own find command,
            # which is useless for us. By redirecting it to the main window,
            # omnivore's find command can be called.
            wx.CallAfter(self.linked_base.editor.control.SetFocus)
            wx.CallAfter(self.ui_action.KeyDown, evt.GetKeyCode(),evt.GetModifiers())
            wx.CallAfter(self.ui_action.KeyUp, evt.GetKeyCode(),evt.GetModifiers())
        else:
            evt.Skip()

    def on_left_down(self, event):
        item = self.HitTest(event.GetPosition())
        if item >= 0:
            selected = self.GetSelection()
            if selected != item:
                self.process_segment_change(item)
        event.Skip()

    def on_click(self, event):
        # BUG: doesn't seem to get called when selecting a segment, using the
        # comments sidebar to jump to another segment, then attempting to
        # select that previous segment. This function never gets called in
        # that case, so I had to add the check on EVT_LEFT_DOWN
        selected = event.GetExtraLong()
        if selected:
            self.process_segment_change(event.GetSelection())
        event.Skip()

    def process_segment_change(self, index):
        segment_uuid = self.index_to_segment_uuid[index]
        if segment_uuid != self.linked_base.segment_uuid:
            wx.CallAfter(self.linked_base.view_segment_uuid, segment_uuid)

    def on_dclick(self, event):
        event.Skip()

    def on_popup(self, event):
        pos = event.GetPosition()
        selected = self.HitTest(pos)
        if selected == -1:
            event.Skip()
            return
        uuid = self.index_to_segment_uuid[selected]
        e = self.linked_base.editor
        # d = e.document

        # # include disabled action showing the name of the segment clicked upon
        # # because it may be different than the selected item
        # name = segment.name
        # if not name:
        #     name = str(segment)
        # actions = [
        #     Action(name=name, task=t, enabled=False),
        #     None,
        #     ]
        # if selected > 0:
        #     actions.append(SelectSegmentInAllAction(segment_uuid=uuid, task=t))
        #     actions.append(ParseSubSegmentsAction(segment_uuid=uuid, task=t))
        #     actions.append(SetSegmentOriginAction(segment_uuid=uuid, task=t))
        #     actions.append(DeleteUserSegmentAction(segment_uuid=uuid, task=t))
        #     actions.append(None)
        # savers = e.get_extra_segment_savers(segment)
        # savers.extend(segment.savers)
        # for saver in savers:
        #     action = SaveSegmentAsFormatAction(saver=saver, segment_uuid=selected, task=t, name="Save as %s" % saver.export_data_name)
        #     actions.append(action)
        # if actions:
        #     e.popup_context_menu_from_actions(self, actions)

    def on_tooltip(self, evt):
        pos = evt.GetPosition()
        selected = self.HitTest(pos)
        if selected >= 0:
            segment = self.linked_base.document.segments[selected]
            self.linked_base.editor.frame.status_message(segment.verbose_info)
        else:
            self.linked_base.editor.frame.status_message("")
        evt.Skip()

    def ensure_visible(self, segment):
        d = self.task.active_editor.document
        index = d.find_segment_index(segment)
        self.EnsureVisible(index)

    def on_key_down(self, evt):
        key = evt.GetKeyCode()
        log.debug("evt=%s, key=%s" % (evt, key))
        moved = False
        index = self.GetSelection()
        if key == wx.WXK_RETURN or key == wx.WXK_TAB:
            self.process_segment_change(index)
        elif key == wx.WXK_UP:
            index = max(index - 1, 0)
            moved = True
        elif key == wx.WXK_DOWN:
            index = min(index + 1, len(self.index_to_segment_uuid) - 1)
            moved = True
        # elif key == wx.WXK_PAGEUP:
        #     index = self.table.get_page_index(e.caret_index, e.segment.page_size, -1, self)
        #     moved = True
        # elif key == wx.WXK_PAGEDOWN:
        #     index = self.table.get_page_index(e.caret_index, e.segment.page_size, 1, self)
        #     moved = True
        else:
            evt.Skip()
        # if True:
        #     evt.Skip()
        #     return
        if moved:
            self.SetSelection(index)
            self.process_segment_change(index)

    def on_key_up(self, evt):
        # Return key not sent through to EVT_CHAR, EVT_CHAR_HOOK or
        # EVT_KEY_DOWN events in a ListBox. This is the only event handler
        # that catches a Return.
        key = evt.GetKeyCode()
        log.debug("key down: %s" % key)
        index = self.GetSelection()
        if key == wx.WXK_RETURN:
            self.task.on_hide_minibuffer_or_cancel(evt)
        evt.Skip()

    def move_caret(self, segment_uuid):
        try:
            index = self.segment_uuid_to_index[segment_uuid]
        except KeyError:
            return
        self.SetSelection(index)



import numpy as np

from atrip import Segment


class AnticFontSegment(Segment):
    def __init__(self, *args, **kwargs):
        Segment.__init__(self, *args, **kwargs)
        if np.alen(self.data) != 1024:
            raise RuntimeError("ANTIC Fonts must be 1024 bytes; have %d bytes" % (np.alen(self.data)))

    @property
    def antic_font(self):
        font = {
            'name': self.name,
            'char_w': 8,
            'char_h': 8,
            'np_data': self.data,
            }
        return font



import numpy as np

from .. import errors
from ..compressor import Compressor

import logging
# logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)
# log.setLevel(logging.DEBUG)


class DCMCompressor(Compressor):
    compression_algorithm = "dcm"

    valid_densities = {
        0: (720, 128),
        1: (720, 256),
        2: (1040, 128),
    }

    def get_next(self):
        try:
            data = self.raw[self.index]
        except IndexError:
            raise errors.InvalidAlgorithm("Incomplete DCM file")
        else:
            self.index += 1
        return data

    def calc_unpacked_data(self, data):
        self.sector_size = 0
        self.num_sectors = 0
        self.current_sector = 0
        self.index = 0
        self.count = len(data)
        self.raw = data
        self.output = np.zeros(200000, dtype=np.uint8)  # max is 1 DD image
        self.current = np.zeros(256, dtype=np.uint8)
        expected_pass = 1
        last_pass = False
        while not last_pass:
            archive_type = self.get_next()
            log.debug(f"index {self.index-1}: archive_type={archive_type:02x}")
            if archive_type == 0xf9 or archive_type == 0xfa:
                archive_flags = self.get_next()
                pass_num = archive_flags & 0x1f
                last_pass = bool(archive_flags & 0x80)
                log.debug(f"index {self.index-1}: pass number={pass_num} last={last_pass}")
                if archive_flags & 0x1f != expected_pass:
                    if archive_type == 0xf9:
                        raise errors.InvalidAlgorithm("DCM multi-file archive combined in the wrong order")
                    else:
                        raise errors.InvalidAlgorithm("Expected pass one of DCM archive first")
                density_flag = (archive_flags >> 5) & 3
                try:
                    self.num_sectors, self.sector_size = self.valid_densities[density_flag]
                except KeyError:
                    raise errors.InvalidAlgorithm(f"Unsupported density flag {density_flag} in DCM")
                log.debug(f"sectors: {self.num_sectors}x{self.sector_size}B")
            else:
                raise errors.InvalidAlgorithm("Not a DCM file")
            self.get_current_sector()

            while True:
                block_type = self.get_next()
                log.debug(f"index {self.index-1}: processing sector {self.current_sector}, type=${block_type & 0x7f:02x}")

                if block_type == 0x45:
                    # pass end, start next pass
                    expected_pass = (expected_pass + 1) % 32
                    log.debug(f"pass end; next pass should be {expected_pass}")
                    break
                try:
                    func = self.decode_block_type_func[block_type & 0x7f]
                except KeyError:
                    if block_type == 0xfa or block_type == 0xf9:
                        raise errors.InvalidAlgorithm(f"Found section start byte but previous section never ended")
                    else:
                        raise errors.InvalidAlgorithm(f"Unsupported block type {block_type} in DCM")
                func(self)
                self.copy_current_to_sector()

                if block_type > 0x80:
                    self.current_sector += 1
                else:
                    self.get_current_sector()

        return self.output[:self.num_sectors * self.sector_size].tobytes()

    def get_current_sector(self):
        lo = self.get_next()
        hi = self.get_next()
        self.current_sector = hi * 256 + lo
        log.debug(f"index {self.index-2}: found sector {self.current_sector}")

    def copy_current_to_sector(self):
        pos = (self.current_sector - 1) * self.sector_size
        self.output[pos:pos + self.sector_size] = self.current[:self.sector_size]

    def decode_41(self):
        """Change beginning of sector"""
        index = self.get_next()
        while index >= 0:
            self.current[index] = self.get_next()
            index -= 1

    def decode_42(self):
        """DOS sector?"""
        self.current[0:124] = self.get_next()
        self.current[124] = self.get_next()
        self.current[125] = self.get_next()
        self.current[126] = self.get_next()
        self.current[127] = self.get_next()

    def decode_43(self):
        """Run-length encoded block"""
        index = 0

        while index < self.sector_size:
            # 1. starts with copying a string verbatim: find end
            end = self.get_next()
            if index > 0 and end == 0:
                end = 256

            # 2: copy bytes verbatim until end offset
            log.debug(f"0x43: copying verbatim {index}-{end}")
            while index < end:
                self.current[index] = self.get_next()
                index += 1

            if index < self.sector_size:
                # 3: run-length encoding
                end = self.get_next()
                if index > 0 and end == 0:
                    end = 256
                fill_byte = self.get_next()
                log.debug(f"0x43: rle: ${fill_byte:02x} {index}-{end}")
                while index < end:
                    self.current[index] = fill_byte
                    index += 1

    def decode_44(self):
        """Change the end of the sector"""
        index = self.get_next()
        while index < self.sector_size:
            self.current[index] = self.get_next()
            index += 1

    def decode_46(self):
        """Same as last sector"""
        return

    def decode_47(self):
        """Uncompressed"""
        log.debug(f"index {self.index}-{self.index+self.sector_size}: uncompressed sector")
        index = 0
        while index < self.sector_size:
            self.current[index] = self.get_next()
            index += 1


    decode_block_type_func = {
        0x41: decode_41,
        0x42: decode_42,
        0x43: decode_43,
        0x44: decode_44,
        0x46: decode_46,
        0x47: decode_47,
    }

    #### compression

    def init_packing(self, media):
        try:
            self.sector_size = media.sector_size
            self.num_sectors = media.num_sectors
        except AttributeError:
            raise errors.InvalidMediaSize("DCM Compressor only works with disk images")
        s = (self.num_sectors, self.sector_size)
        for density_flag, size in self.valid_densities.items():
            if s == size:
                self.density_flag = density_flag
                break
        else:
            raise errors.InvalidMediaSize("DCM Compressor only works with standard size Atari disk images")
        self.index = 0
        self.output = np.zeros(200000, dtype=np.uint8)  # max is 1 DD image
        self.current = np.zeros(256, dtype=np.uint8)
        self.previous = np.zeros(256, dtype=np.uint8)
        self.pass_buffer = np.zeros(0x6500, dtype=np.uint8)
        self.pass_buffer_index = 0
        self.record_start_index = 0
        self.pass_number = 1

    def put_byte(self, value):
        self.pass_buffer[self.pass_buffer_index] = value
        self.pass_buffer_index += 1

    def calc_packed_data(self, byte_data, media, block_restrictions=None):
        self.init_packing(media)
        output_index = 0
        current_sector = 1
        previous_sector = 0

        while current_sector <= self.num_sectors:
            log.debug(f"dcm: starting pass {self.pass_number} at sector {current_sector}")
            self.encode_fa(current_sector)
            first_sector_in_pass = 0

            while self.pass_buffer_index < 0x5e00:
                if current_sector > self.num_sectors:
                    break
                pos, size = media.get_index_of_sector(current_sector)
                self.current[:size] = media[pos:pos + size]
                if np.count_nonzero(self.current[:size]) == 0:  # empty
                    log.debug(f"dcm: skipping empty sector {current_sector}")
                    current_sector += 1
                else:
                    if first_sector_in_pass == 0:
                        first_sector_in_pass = current_sector
                        previous_sector = current_sector
                    if current_sector - previous_sector > 1:
                        # save the first non-blank sector if there is a run of
                        # blank sectors
                        self.encode_sector(current_sector)
                    else:
                        self.encode_implicit_next_sector()
                    log.debug(f"dcm: encoding sector {current_sector} at buffer index {self.pass_buffer_index}")

                    self.encode_best(current_sector == first_sector_in_pass, block_restrictions)

                    # save current sector data so it can be compared
                    self.previous[:size] = self.current[:size]
                    previous_sector = current_sector
                    current_sector += 1

            log.debug(f"dcm: filled buffer for pass {self.pass_number}")

            # force next pass to not attempt to read a sector number
            self.encode_implicit_next_sector()

            self.encode_45()

            pass_length = self.pass_buffer_index

            # rerecord FA block to show if it is the final pass
            last_pass = current_sector > self.num_sectors
            self.encode_fa(first_sector_in_pass, last_pass)

            self.output[output_index:output_index + pass_length] = self.pass_buffer[:pass_length]
            output_index += pass_length
            self.pass_number += 1

        return self.output[:output_index].tobytes()

    def encode_sector(self, sector):
        self.put_byte(sector & 0xff)
        self.put_byte(sector >> 8)

    def encode_implicit_next_sector(self):
        self.pass_buffer[self.record_start_index] |= 0x80

    def encode_fa(self, sector, last_pass=False):
        self.record_start_index = 0
        self.pass_buffer_index = 0
        self.put_byte(0xfa)
        flag = self.density_flag << 5 | self.pass_number & 0x1f
        if last_pass:
            flag |= 0x80
        self.put_byte(flag)
        self.encode_sector(sector)

    def encode_45(self):
        self.record_start_index = self.pass_buffer_index
        self.put_byte(0x45)

    def encode_best(self, first=False, block_restrictions=None):
        # NOTE: 0x46, 0x47 always allowed
        if block_restrictions is None:
            allowed_blocks = set([0x41, 0x42, 0x43, 0x44])
        else:
            allowed_blocks = block_restrictions
        size = self.sector_size
        best_size = size
        best_block = 0x47
        encoder_arg = 0

        if not first:
            same = self.current[:size] - self.previous[:size]
            where_diff = np.where(same != 0)[0]
            # print("same", same)
            # print("where_diff", where_diff)
            try:
                first_diff = where_diff[0]
            except IndexError:
                # no differences!
                best_block = 0x46
                best_size = 0
            else:
                last_diff = where_diff[-1]
                len_41 = last_diff + 1
                if len_41 < best_size and 0x41 in allowed_blocks:
                    # print(f"41: {last_diff}, {where_diff}")
                    best_size = len_41
                    best_block = 0x41
                    encoder_arg = last_diff

                len_44 = size - first_diff + 1
                if len_44 < best_size and 0x44 in allowed_blocks:
                    # print("current", self.current[:size])
                    # print("previous", self.previous[:size])
                    # print("same", same)
                    # print("same", self.current[:size] - self.previous[:size])
                    # print("where_diff", where_diff)
                    # print(f"44: {first_diff}, {where_diff}")
                    best_size = len_44
                    best_block = 0x44
                    encoder_arg = first_diff

        same = np.diff(self.current[:size] - self.current[0])
        where_diff = np.where(same != 0)[0]
        try:
            first_diff = where_diff[0] + 1
        except IndexError:
            # no differences anywhere
            first_diff = self.sector_size
        if first_diff > 123 and best_size > 6 and 0x42 in allowed_blocks:
            best_block = 0x42
            best_size = 6
        if 0x43 in allowed_blocks:
            len_43, groups_43 = self.prepare_43()
            if len_43 < best_size and 0x43 in allowed_blocks:
                best_size = len_43
                best_block = 0x43
                encoder_arg = groups_43

        log.debug(f"dcm: encoding as block type ${best_block:x}, size={best_size}")
        func = self.encode_block_type_func[best_block]
        func(self, encoder_arg)

    def encode_41(self, index):
        """Change beginning of sector"""
        self.record_start_index = self.pass_buffer_index
        self.put_byte(0x41)
        self.put_byte(index)
        while index >= 0:
            self.put_byte(self.current[index])
            index -= 1

    def encode_42(self, unused):
        """DOS sector? Single density only, apparently"""
        self.record_start_index = self.pass_buffer_index
        self.put_byte(0x42)
        self.put_byte(self.current[123])
        self.put_byte(self.current[124])
        self.put_byte(self.current[125])
        self.put_byte(self.current[126])
        self.put_byte(self.current[127])

    def prepare_43(self):
        size = self.sector_size

        # find where the values are the same
        d = np.diff(self.current[:size])
        # [  1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
        #    1,   1,   1,   1,   1,   1,  -9,   0,   0,   0,   0,   0,   0,
        #    0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        #   30,   1,   1,   1,   1,   1,   1,   1,   1,   1,  50,   0,   0,
        #    0,   0,   0,   0,   0,   0,   0, -95,   0,   0,   0,   0,   0,
        #    0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        #    0,  76,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
        #    1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
        #    1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
        #    1,   1,   1,   1,   1,   1,   1,   1,   1,   1]

        # The index before the first zero in a group of zeros is the start of a
        # group of the same values. This gives a list of indexes of the groups
        same = np.where(d == 0)[0]
        # [20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36,
        #  37, 38, 50, 51, 52, 53, 54, 55, 56, 57, 58, 60, 61, 62, 63, 64, 65,
        #  66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78]

        # If no consecutive bytes have the same value, abort!
        if len(same) == 0:
            return 100000, None

        # The changes in same are the breaks in groups:
        changes=np.diff(same)
        # [ 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
        #   1, 12,  1,  1,  1,  1,  1,  1,  1,  1,  2,  1,  1,  1,  1,  1,  1,
        #   1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1]

        # each 1 represents the same value as the next item in the array, so
        # values > 1 are where there are gaps, i.e. the start of a new group
        starts = np.where(changes > 1)[0] + 1
        # [19, 28]
        ends = starts - 1
        # [18, 27]

        # but it doesn't include the start of the first group or the end of the
        # last group
        starts = list(starts)
        starts[0:0] = [0]
        ends = list(ends)
        ends.append(-1)

        rle_starts = same[starts]
        # [20, 50, 60]
        rle_ends = same[ends] + 1 + 1  # inclusive, so need extra to form slice
        # [40, 60, 80]
        rle_groups = list(zip(rle_starts, rle_ends))

        # start with verbatim copy, then alternate with rle
        length = 0
        index = 0
        for start, end in rle_groups:
            length += 1 + start - index  # verbatim
            index = start
            if index < size:
                length += 2  # rle blocks always encoded in 2 bytes
            index = end
        if index < size:
            length += 1 + size - index
            rle_groups.append((size, size))
        return length, rle_groups

    def encode_43(self, groups):
        """Run-length encoded block"""
        self.record_start_index = self.pass_buffer_index
        self.put_byte(0x43)
        index = 0
        size = self.sector_size
        for rle_start, rle_end in groups:
            self.put_byte(rle_start)
            while index < rle_start:
                self.put_byte(self.current[index])
                index += 1

            if index < size:
                self.put_byte(rle_end)
                self.put_byte(self.current[index])
                index += rle_end - rle_start

    def encode_44(self, index):
        """Change beginning of sector"""
        self.record_start_index = self.pass_buffer_index
        self.put_byte(0x44)
        self.put_byte(index)
        while index < self.sector_size:
            self.put_byte(self.current[index])
            index += 1

    def encode_46(self, unused):
        self.record_start_index = self.pass_buffer_index
        self.put_byte(0x46)

    def encode_47(self, index):
        """Uncompressed"""
        self.record_start_index = self.pass_buffer_index
        self.put_byte(0x47)
        index = 0
        while index < self.sector_size:
            self.put_byte(self.current[index])
            index += 1


    encode_block_type_func = {
        0x41: encode_41,
        0x42: encode_42,
        0x43: encode_43,
        0x44: encode_44,
        0x46: encode_46,
        0x47: encode_47,
    }

# Notes:
#
# 
# >>> current=np.arange(128, dtype=np.uint8)
# >>> previous=np.arange(128, dtype=np.uint8)
# >>> previous[80:100]=0xff
# >>> d=np.diff(current-previous)
# >>> d
# array([  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
#          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
#          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
#          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
#          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
#          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
#          0,  81,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
#          1,   1,   1,   1,   1,   1,   1,   1, 156,   0,   0,   0,   0,
#          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
#          0,   0,   0,   0,   0,   0,   0,   0,   0,   0], dtype=uint8)
# >>> np.where(d!=0)[0]
# array([79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95,
#        96, 97, 98, 99])
# >>> w=np.where(d!=0)[0]
# >>> w
# array([79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95,
#        96, 97, 98, 99])
# >>> first_diff = w[0] + 1
# >>> last_diff = w[-1] + 1
# >>> previous[0:first_diff]
# array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16,
#        17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33,
#        34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50,
#        51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67,
#        68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79], dtype=uint8)
# >>> previous[first_diff:last_diff]
# array([255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
#        255, 255, 255, 255, 255, 255, 255], dtype=uint8)
# >>> previous[last_diff:]
# array([100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112,
#        113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125,
#        126, 127], dtype=uint8)
# >>> 



# RLE notes:
#
# >>> a=np.arange(128)
# >>> a[20:40]=10
# >>> a[50:60]=99
# >>> a[60:80]=4
# >>> a
# array([  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12,
#         13,  14,  15,  16,  17,  18,  19,  10,  10,  10,  10,  10,  10,
#         10,  10,  10,  10,  10,  10,  10,  10,  10,  10,  10,  10,  10,
#         10,  40,  41,  42,  43,  44,  45,  46,  47,  48,  49,  99,  99,
#         99,  99,  99,  99,  99,  99,  99,  99,   4,   4,   4,   4,   4,
#          4,   4,   4,   4,   4,   4,   4,   4,   4,   4,   4,   4,   4,
#          4,   4,  80,  81,  82,  83,  84,  85,  86,  87,  88,  89,  90,
#         91,  92,  93,  94,  95,  96,  97,  98,  99, 100, 101, 102, 103,
#        104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116,
#        117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127])

# >>> d=np.diff(a)
# >>> d
# array([  1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
#          1,   1,   1,   1,   1,   1,  -9,   0,   0,   0,   0,   0,   0,
#          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
#         30,   1,   1,   1,   1,   1,   1,   1,   1,   1,  50,   0,   0,
#          0,   0,   0,   0,   0,   0,   0, -95,   0,   0,   0,   0,   0,
#          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
#          0,  76,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
#          1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
#          1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
#          1,   1,   1,   1,   1,   1,   1,   1,   1,   1])
# >>> 
#
# The index before the first zero in a group of zeros is the start of a group
# of the same values. This gives a list of indexes of the groups:
#
# >>> same=np.where(d==0)[0]
# >>> same
# array([20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36,
#        37, 38, 50, 51, 52, 53, 54, 55, 56, 57, 58, 60, 61, 62, 63, 64, 65,
#        66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78])
#
# and where it is different:
#
# >>> changes=np.diff(same)
# >>> changes
# array([ 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
#         1, 12,  1,  1,  1,  1,  1,  1,  1,  1,  2,  1,  1,  1,  1,  1,  1,
#         1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1])
#
# should be the start of a new group.
#
# >>> starts=np.where(changes > 1)[0] + 1
# >>> starts
# array([19, 28])
# >>> same[0:19]
# array([20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36,
#        37, 38])
# >>> same[19:28]
# array([50, 51, 52, 53, 54, 55, 56, 57, 58])
# >>> same[28:]
# array([60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76,
#        77, 78])



import numpy as np

from . import errors
from .diskimages import DiskImageBase, BaseHeader, Bootable
from .segments import SegmentData, EmptySegment, ObjSegment, RawSectorsSegment, DefaultSegment, SegmentedFileSegment, SegmentSaver
from .style_bits import get_style_bits
from .utils import *
from .executables import get_xex

import logging
log = logging.getLogger(__name__)
try:  # Expensive debugging
    _xd = _expensive_debugging
except NameError:
    _xd = False


class AtariDosWriteableSector(WriteableSector):
    @property
    def next_sector_num(self):
        return self._next_sector_num

    @next_sector_num.setter
    def next_sector_num(self, value):
        self._next_sector_num = value
        index = self.sector_size - 3
        hi, lo = divmod(value, 256)
        self.data[index + 0] = (self.file_num << 2) | (hi & 0x03)
        self.data[index + 1] = lo
        self.data[index + 2] = self.used
        if _xd: log.debug("sector metadata for %d: %s" % (self._sector_num, self.data[index:index + 3]))
        # file number will be added later when known.


class AtariDosVTOC(VTOC):
    def parse_segments(self, segments):
        self.vtoc1 = segments[0].data
        bits = np.unpackbits(self.vtoc1[0x0a:0x64])
        self.sector_map[0:720] = bits
        if _xd: log.debug("vtoc before:\n%s" % str(self))

    def calc_bitmap(self):
        if _xd: log.debug("vtoc after:\n%s" % str(self))
        packed = np.packbits(self.sector_map[0:720])
        self.vtoc1[0x0a:0x64] = packed
        s = WriteableSector(self.sector_size, self.vtoc1)
        s.sector_num = 360
        self.sectors.append(s)


class AtariDosDirectory(Directory):
    @property
    def dirent_class(self):
        return AtariDosDirent

    def encode_empty(self):
        return np.zeros([16], dtype=np.uint8)

    def encode_dirent(self, dirent):
        data = dirent.encode_dirent()
        if _xd: log.debug("encoded dirent: %s" % data)
        return data

    def set_sector_numbers(self, image):
        num = 361
        for sector in self.sectors:
            sector.sector_num = num
            num += 1


class AtariDosDirent(Dirent):
    # ATR Dirent structure described at http://atari.kensclassics.org/dos.htm
    format = np.dtype([
        ('FLAG', 'u1'),
        ('COUNT', '<u2'),
        ('START', '<u2'),
        ('NAME','S8'),
        ('EXT','S3'),
        ])

    def __init__(self, image, file_num=0, bytes=None):
        Dirent.__init__(self, file_num)
        self.flag = 0
        self.opened_output = False
        self.dos_2 = False
        self.mydos = False
        self.is_dir = False
        self.locked = False
        self.in_use = False
        self.deleted = False
        self.num_sectors = 0
        self.starting_sector = 0
        self.basename = b''
        self.ext = b''
        self.is_sane = True
        self.current_sector = 0
        self.current_read = 0
        self.sectors_seen = None
        self.parse_raw_dirent(image, bytes)

    def __str__(self):
        return "File #%-2d (%s) %03d %-8s%-3s  %03d" % (self.file_num, self.summary, self.starting_sector, self.basename.decode("latin1"), self.ext.decode("latin1"), self.num_sectors)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.filename == other.filename and self.starting_sector == other.starting_sector and self.num_sectors == other.num_sectors

    @property
    def filename(self):
        ext = (b'.' + self.ext) if self.ext else b''
        return (self.basename + ext).decode('latin1')

    @property
    def summary(self):
        output = "o" if self.opened_output else "."
        dos2 = "2" if self.dos_2 else "."
        mydos = "m" if self.mydos else "."
        in_use = "u" if self.in_use else "."
        deleted = "d" if self.deleted else "."
        locked = "*" if self.locked else " "
        flags = "%s%s%s%s%s%s" % (output, dos2, mydos, in_use, deleted, locked)
        return flags

    @property
    def verbose_info(self):
        flags = []
        if self.opened_output: flags.append("OUT")
        if self.dos_2: flags.append("DOS2")
        if self.mydos: flags.append("MYDOS")
        if self.in_use: flags.append("IN_USE")
        if self.deleted: flags.append("DEL")
        if self.locked: flags.append("LOCK")
        return "flags=[%s]" % ", ".join(flags)

    def extra_metadata(self, image):
        return self.verbose_info

    def parse_raw_dirent(self, image, data):
        if data is None:
            return
        values = data.view(dtype=self.format)[0]
        flag = values[0]
        self.flag = flag
        self.opened_output = (flag&0x01) > 0
        self.dos_2 = (flag&0x02) > 0
        self.mydos = (flag&0x04) > 0
        self.is_dir = (flag&0x10) > 0
        self.locked = (flag&0x20) > 0
        self.in_use = (flag&0x40) > 0
        self.deleted = (flag&0x80) > 0
        self.num_sectors = int(values[1])
        self.starting_sector = int(values[2])
        self.basename = bytes(values[3]).rstrip()
        self.ext = bytes(values[4]).rstrip()
        self.is_sane = self.sanity_check(image)

    def encode_dirent(self):
        data = np.zeros([self.format.itemsize], dtype=np.uint8)
        values = data.view(dtype=self.format)[0]
        flag = (1 * int(self.opened_output)) | (2 * int(self.dos_2)) | (4 * int(self.mydos)) | (0x10 * int(self.is_dir)) | (0x20 * int(self.locked)) | (0x40 * int(self.in_use)) | (0x80 * int(self.deleted))
        values[0] = flag
        values[1] = self.num_sectors
        values[2] = self.starting_sector
        values[3] = self.basename
        values[4] = self.ext
        return data

    def mark_deleted(self):
        self.deleted = True
        self.in_use = False

    def update_sector_info(self, sector_list):
        self.num_sectors = sector_list.num_sectors
        self.starting_sector = sector_list.first_sector

    def add_metadata_sectors(self, vtoc, sector_list, header):
        # no extra sectors are needed for an Atari DOS file; the links to the
        # next sector is contained in the sector.
        pass

    def sanity_check(self, image):
        if not self.in_use:
            return True
        if not image.header.sector_is_valid(self.starting_sector):
            return False
        if self.num_sectors < 0 or self.num_sectors > image.header.max_sectors:
            return False
        return True

    def get_sectors_in_vtoc(self, image):
        sector_list = BaseSectorList(image.header)
        self.start_read(image)
        while True:
            sector = WriteableSector(image.header.sector_size, None, self.current_sector)
            sector_list.append(sector)
            _, last, _, _ = self.read_sector(image)
            if last:
                break
        return sector_list

    def start_read(self, image):
        if not self.is_sane:
            raise errors.InvalidDirent("Invalid directory entry '%s'" % str(self))
        self.current_sector = self.starting_sector
        self.current_read = self.num_sectors
        self.sectors_seen = set()

    def read_sector(self, image):
        raw, pos, size = image.get_raw_bytes(self.current_sector)
        bytes, num_data_bytes = self.process_raw_sector(image, raw)
        return bytes, self.current_sector == 0, pos, num_data_bytes

    def process_raw_sector(self, image, raw):
        file_num = raw[-3] >> 2
        if file_num != self.file_num:
            raise errors.FileNumberMismatchError164("Expecting file %d, found %d" % (self.file_num, file_num))
        self.sectors_seen.add(self.current_sector)
        next_sector = ((raw[-3] & 0x3) << 8) + raw[-2]
        if next_sector in self.sectors_seen:
            raise errors.InvalidFile("Bad sector pointer data: attempting to reread sector %d" % next_sector)
        self.current_sector = next_sector
        num_bytes = raw[-1]
        return raw[0:num_bytes], num_bytes

    def set_values(self, filename, filetype, index):
        if type(filename) is not bytes:
            filename = filename.encode("latin1")
        if b'.' in filename:
            filename, ext = filename.split(b'.', 1)
        else:
            ext = b'   '
        self.basename = b'%-8s' % filename[0:8]
        self.ext = ext
        self.file_num = index
        self.dos_2 = True
        self.in_use = True
        if _xd: log.debug("set_values: %s" % self)


class MydosDirent(AtariDosDirent):
    def process_raw_sector(self, image, raw):
        # No file number stored in the sector data; two full bytes available
        # for next sector
        self.current_sector = (raw[-3] << 8) + raw[-2]
        num_bytes = raw[-1]
        return raw[0:num_bytes], num_bytes


class XexSegmentSaver(SegmentSaver):
    export_data_name = "Atari 8-bit Executable"
    export_extensions = [".xex"]


class XexContainerSegment(DefaultSegment):
    can_resize_default = True


class XexSegment(ObjSegment):
    savers = [SegmentSaver, XexSegmentSaver]


class RunAddressSegment(ObjSegment):
    # FIXME: defining run_address as a property doesn't work for some reason.
    # @property
    # def run_address(self):
    #     return self.rawdata[0:2].view(dtype="<u2")[0]
    def run_address(self):
        return self.rawdata[0:2].data.view(dtype="<u2")[0]



class AtariDosFile(Bootable):
    """Parse a binary chunk into segments according to the Atari DOS object
    file format.
    
    Ref: http://www.atarimax.com/jindroush.atari.org/afmtexe.html
    """

    def __init__(self, rawdata):
        self.rawdata = rawdata
        self.size = len(rawdata)
        self.segments = []
        self.files = []

    def __str__(self):
        return "\n".join(str(s) for s in self.segments) + "\n"

    def strict_check(self):
        pass

    def relaxed_check(self):
        pass

    def parse_segments(self):
        r = self.rawdata
        b = r.get_data()
        s = r.get_style()
        pos = 0
        style_pos = 0
        first = True
        if _xd: log.debug("Initial parsing: size=%d" % self.size)
        while pos < self.size:
            if pos + 1 < self.size:
                header, = b[pos:pos+2].view(dtype='<u2')
            else:
                self.segments.append(ObjSegment(r[pos:pos + 1], pos, pos + 1, 0, 1, "Incomplete Data"))
                break
            if header == 0xffff:
                # Apparently 0xffff header can appear in any segment, not just
                # the first.  Regardless, it is ignored everywhere.
                pos += 2
                first = False
                continue
            elif first:
                raise errors.InvalidBinaryFile("Object file doesn't start with 0xffff")
            if _xd: log.debug("header parsing: header=0x%x" % header)
            if len(b[pos:pos + 4]) < 4:
                self.segments.append(ObjSegment(r[pos:pos + 4], 0, 0, 0, len(b[pos:pos + 4]), "Short Segment Header"))
                break
            start, end = b[pos:pos + 4].view(dtype='<u2')
            s[style_pos:pos + 4] = get_style_bits(data=True)
            if end < start:
                raise errors.InvalidBinaryFile("Nonsensical start and end addresses")
            count = end - start + 1
            found = len(b[pos + 4:pos + 4 + count])
            if found < count:
                self.segments.append(ObjSegment(r[pos + 4:pos + 4 + count], pos, pos + 4, start, end, "Incomplete Data"))
                break
            if start == 0x2e0:
                segment_cls = RunAddressSegment
            else:
                segment_cls = ObjSegment
            self.segments.append(segment_cls(r[pos + 4:pos + 4 + count], pos, pos + 4, start, end))
            pos += 4 + count
            style_pos = pos


class AtrHeader(BaseHeader):
    sector_class = AtariDosWriteableSector

    # ATR Format described in http://www.atarimax.com/jindroush.atari.org/afmtatr.html
    format = np.dtype([
        ('wMagic', '<u2'),
        ('wPars', '<u2'),
        ('wSecSize', '<u2'),
        ('btParsHigh', 'u1'),
        ('dwCRC','<u4'),
        ('unused','<u4'),
        ('btFlags','u1'),
        ])
    file_format = "ATR"

    def __init__(self, bytes=None, sector_size=128, initial_sectors=3, create=False):
        BaseHeader.__init__(self, sector_size, initial_sectors, 360, 1)
        if create:
            self.header_offset = 16
            self.check_size(0)
        if bytes is None:
            return

        if len(bytes) == 16:
            values = bytes.view(dtype=self.format)[0]
            if values[0] != 0x296:
                raise errors.InvalidAtrHeader("no ATR header magic value")
            self.image_size = (int(values[3]) * 256 * 256 + int(values[1])) * 16
            self.sector_size = int(values[2])
            self.crc = int(values[4])
            self.unused = int(values[5])
            self.flags = int(values[6])
            self.header_offset = 16
        else:
            raise errors.InvalidAtrHeader("incorrect AHC header size of %d" % len(bytes))

    def __str__(self):
        return "%s Disk Image (size=%d (%dx%dB), crc=%d flags=%d unused=%d)" % (self.file_format, self.image_size, self.max_sectors, self.sector_size, self.crc, self.flags, self.unused)

    def encode(self, raw):
        values = raw.view(dtype=self.format)[0]
        values[0] = 0x296
        paragraphs = self.image_size // 16
        parshigh, pars = divmod(paragraphs, 256*256)
        values[1] = pars
        values[2] = self.sector_size
        values[3] = parshigh
        values[4] = self.crc
        values[5] = self.unused
        values[6] = self.flags
        return raw

    def check_size(self, size):
        if size == 92160 or size == 92176:
            self.image_size = 92160
            self.sector_size = 128
            self.initial_sector_size = 0
            self.num_initial_sectors = 0
        elif size == 184320 or size == 184336:
            self.image_size = 184320
            self.sector_size = 256
            self.initial_sector_size = 0
            self.num_initial_sectors = 0
        elif size == 183936 or size == 183952:
            self.image_size = 183936
            self.sector_size = 256
            self.initial_sector_size = 128
            self.num_initial_sectors = 3
        else:
            self.image_size = size
        self.first_vtoc = 360
        self.num_vtoc = 1
        self.first_directory = 361
        self.num_directory = 8
        self.tracks_per_disk = 40
        self.sectors_per_track = 18
        self.payload_bytes = self.sector_size - 3
        initial_bytes = self.initial_sector_size * self.num_initial_sectors
        self.max_sectors = ((self.image_size - initial_bytes) // self.sector_size) + self.num_initial_sectors

    def get_pos(self, sector):
        if not self.sector_is_valid(sector):
            raise errors.ByteNotInFile166("Sector %d out of range" % sector)
        if sector <= self.num_initial_sectors:
            pos = self.num_initial_sectors * (sector - 1)
            size = self.initial_sector_size
        else:
            pos = self.num_initial_sectors * self.initial_sector_size + (sector - 1 - self.num_initial_sectors) * self.sector_size
            size = self.sector_size
        pos += self.header_offset
        return pos, size

    def strict_check(self, image):
        size = len(image)
        if self.header_offset == 16 or size in [92176, 133136, 184336, 183952]:
            return
        raise errors.InvalidDiskImage("Uncommon size of ATR file")


class XfdHeader(AtrHeader):
    file_format = "XFD"

    def __str__(self):
        return "%s Disk Image (size=%d (%dx%dB)" % (self.file_format, self.image_size, self.max_sectors, self.sector_size)

    def __len__(self):
        return 0

    def to_array(self):
        raw = np.zeros([0], dtype=np.uint8)
        return raw

    def strict_check(self, image):
        size = len(image)
        if size in [92160, 133120, 183936, 184320]:
            return
        raise errors.InvalidDiskImage("Uncommon size of XFD file")


class AtariDosDiskImage(DiskImageBase):
    default_executable_extension = "XEX"

    def __init__(self, *args, **kwargs):
        self.first_vtoc = 360
        self.num_vtoc = 1
        self.vtoc2 = 0
        self.first_data_after_vtoc = 369
        DiskImageBase.__init__(self, *args, **kwargs)

    @property
    def writeable_sector_class(self):
        return AtariDosWriteableSector

    @property
    def vtoc_class(self):
        return AtariDosVTOC

    @property
    def directory_class(self):
        return AtariDosDirectory

    def __str__(self):
        return "%s Atari DOS Format: %d usable sectors (%d free), %d files" % (self.header, self.total_sectors, self.unused_sectors, len(self.files))

    @classmethod
    def new_header(cls, diskimage, format="ATR"):
        if format.lower() == "atr":
            header = AtrHeader(create=True)
            header.check_size(diskimage.size)
        else:
            raise RuntimeError("Unknown header type %s" % format)
        return header

    def as_new_format(self, format="ATR"):
        """ Create a new disk image in the specified format
        """
        first_data = len(self.header)
        raw = self.rawdata[first_data:]
        data = add_atr_header(raw)
        newraw = SegmentData(data)
        image = self.__class__(newraw)
        return image

    vtoc_type = np.dtype([
        ('code', 'u1'),
        ('total','<u2'),
        ('unused','<u2'),
        ])

    def read_header(self):
        bytes = self.bytes[0:16]
        try:
            self.header = AtrHeader(bytes)
        except errors.InvalidAtrHeader:
            self.header = XfdHeader()

    def calc_vtoc_code(self):
        # From AA post: http://atariage.com/forums/topic/179868-mydos-vtoc-size/
        num = 1 + (self.total_sectors + 80) // (self.header.sector_size * 8)
        if self.header.sector_size == 128:
            if num == 1:
                code = 2
            else:
                if num & 1:
                    num += 1
                code = ((num + 1) // 2) + 2
        else:
            if self.total_sectors < 1024:
                code = 2
            else:
                code = 2 + num
        return code

    def get_vtoc(self):
        data, style = self.get_sectors(360)
        values = data[0:5].view(dtype=self.vtoc_type)[0]
        code = values[0]
        if code == 0 or code == 2:
            num = 1
        else:
            num = (code * 2) - 3
        self.first_vtoc = 360 - num + 1
        self.assert_valid_sector(self.first_vtoc)
        self.num_vtoc = num
        if num < 0 or num > self.calc_vtoc_code():
            raise errors.InvalidDiskImage("Invalid number of VTOC sectors: %d" % num)

        self.total_sectors = values[1]
        self.unused_sectors = values[2]
        if self.header.image_size == 133120:
            # enhanced density has 2nd VTOC
            self.vtoc2 = 1024
            data, style = self.get_sectors(self.vtoc2)
            extra_free = data[122:124].view(dtype='<u2')[0]
            self.unused_sectors += extra_free

    def get_directory(self, directory=None):
        dir_bytes, style = self.get_sectors(361, 368)
        i = 0
        num = 0
        files = []
        while i < len(dir_bytes):
            dirent = AtariDosDirent(self, num, dir_bytes[i:i+16])
            if dirent.mydos:
                dirent = MydosDirent(self, num, dir_bytes[i:i+16])

            if dirent.in_use:
                files.append(dirent)
                if not dirent.is_sane:
                    self.all_sane = False
                    log.debug("dirent %d not sane: %s" % (num, dirent))
            elif dirent.flag == 0:
                break
            if directory is not None:
                directory.set(num, dirent)
            i += 16
            num += 1
        self.files = files

    boot_record_type = np.dtype([
        ('BFLAG', 'u1'),
        ('BRCNT', 'u1'),
        ('BLDADR', '<u2'),
        ('BWTARR', '<u2'),
        ('jmp', 'u1'),
        ('XBCONT', '<u2'),
        ('SABYTE', 'u1'),
        ('DRVBYT', 'u1'),
        ('unused', 'u1'),
        ('SASA', '<u2'),
        ('DFSFLG', 'u1'),
        ('DFLINK', '<u2'),
        ('BLDISP', 'u1'),
        ('DFLADR', '<u2'),
        ])

    def get_boot_segments(self):
        data, style = self.get_sectors(360)
        values = data[0:20].view(dtype=self.boot_record_type)[0]
        flag = int(values[0])
        segments = []
        if flag == 0:
            num = int(values[1])
            addr = int(values[2])
            s = self.get_sector_slice(1, num)
            r = self.rawdata[s]
            header = ObjSegment(r[0:20], 0, 0, addr, addr + 20, name="Boot Header")
            sectors = ObjSegment(r, 0, 0, addr, addr + len(r), name="Boot Sectors")
            code = ObjSegment(r[20:], 0, 0, addr + 20, addr + len(r), name="Boot Code")
            segments = [sectors, header, code]
        return segments

    def get_vtoc_segments(self):
        r = self.rawdata
        segments = []
        addr = 0
        start, count = self.get_contiguous_sectors(self.first_vtoc, self.num_vtoc)
        segment = RawSectorsSegment(r[start:start+count], self.first_vtoc, self.num_vtoc, count, 128, 3, self.header.sector_size, name="VTOC")
        segment.style[:] = get_style_bits(data=True)
        segment.set_comment_at(0x00, "Type code")
        segment.set_comment_at(0x01, "Total number of sectors")
        segment.set_comment_at(0x03, "Number of free sectors")
        segment.set_comment_at(0x05, "reserved")
        segment.set_comment_at(0x06, "unused")
        segment.set_comment_at(0x0a, "Sector bit map")
        segment.set_comment_at(0x64, "unused")
        segments.append(segment)
        if self.vtoc2 > 0:
            start, count = self.get_contiguous_sectors(self.vtoc2, 1)
            segment = RawSectorsSegment(r[start:start+count], self.vtoc2, 1, count, 128, 3, self.header.sector_size, name="VTOC2")
            segment.style[:] = get_style_bits(data=True)
            segment.set_comment_at(0x00, "Repeat of sectors 48-719")
            segment.set_comment_at(0x44, "Sector bit map 720-1023")
            segment.set_comment_at(0x7a, "Number of free sectors above 720")
            segment.set_comment_at(0x7c, "unused")
            segments.append(segment)
        return segments

    def get_directory_segments(self):
        r = self.rawdata
        segments = []
        addr = 0
        start, count = self.get_contiguous_sectors(361, 8)
        segment = RawSectorsSegment(r[start:start+count], 361, 8, count, 128, 3, self.header.sector_size, name="Directory")
        segment.style[:] = get_style_bits(data=True)
        index = 0
        for filenum in range(64):
            segment.set_comment_at(index + 0x00, "FILE #%d: Flag" % filenum)
            segment.set_comment_at(index + 0x01, "FILE #%d: Number of sectors in file" % filenum)
            segment.set_comment_at(index + 0x03, "FILE #%d: Starting sector number" % filenum)
            segment.set_comment_at(index + 0x05, "FILE #%d: Filename" % filenum)
            segment.set_comment_at(index + 0x0d, "FILE #%d: Extension" % filenum)
            index += 16
        segments.append(segment)
        return segments

    def get_file_segment(self, dirent):
        byte_order = []
        dirent.start_read(self)
        while True:
            bytes, last, pos, size = dirent.read_sector(self)
            byte_order.extend(list(range(pos, pos + size)))
            if last:
                break
        if len(byte_order) > 0:
            name = "%s %ds@%d" % (dirent.filename, dirent.num_sectors, dirent.starting_sector)
            verbose_name = "%s (%d sectors, first@%d) %s" % (dirent.filename, dirent.num_sectors, dirent.starting_sector, dirent.verbose_info)
            raw = self.rawdata.get_indexed(byte_order)
            segment = DefaultSegment(raw, name=name, verbose_name=verbose_name)
        else:
            segment = EmptySegment(self.rawdata, name=dirent.filename)
        return segment

    def get_file_segments(self):
        segments_in = DiskImageBase.get_file_segments(self)
        segments_out = []
        for segment in segments_in:
            segments_out.append(segment)
            try:
                binary = AtariDosFile(segment.rawdata)
                segments_out.extend(binary.segments)
            except errors.InvalidBinaryFile:
                log.debug("%s not a binary file; skipping segment generation" % str(segment))
        return segments_out


class BootDiskImage(AtariDosDiskImage):
    def __str__(self):
        return "%s Boot Disk" % (self.header)

    def check_size(self):
        if self.header is None:
            return
        start, size = self.header.get_pos(1)
        b = self.bytes
        i = self.header.header_offset
        flag = b[i:i + 2].view(dtype='<u2')[0]
        if flag == 0xffff:
            raise errors.InvalidDiskImage("Appears to be an executable")
        nsec = b[i + 1]
        bload = b[i + 2:i + 4].view(dtype='<u2')[0]

        # Sanity check: number of sectors to be loaded can't be more than the
        # lower 48k of ram because there's no way to bank switch or anything
        # before the boot sectors are finished loading
        max_ram = 0xc000
        max_size = max_ram - bload
        max_sectors = max_size // self.header.sector_size
        if nsec > max_sectors or nsec < 1:
            raise errors.InvalidDiskImage("Number of boot sectors out of range (tried %d, max=%d" % (nsec, max_sectors))
        if bload > (0xc000 - (nsec * self.header.sector_size)):
            raise errors.InvalidDiskImage("Bad boot load address")

    def get_boot_sector_info(self):
        pass

    def get_vtoc(self):
        pass

    def get_directory(self, directory=None):
        pass

    boot_record_type = np.dtype([
        ('BFLAG', 'u1'),
        ('BRCNT', 'u1'),
        ('BLDADR', '<u2'),
        ('BWTARR', '<u2'),
        ])

    def get_boot_segments(self):
        data, style = self.get_sectors(1)
        values = data[0:6].view(dtype=self.boot_record_type)[0]
        flag = int(values[0])
        segments = []
        if flag == 0:
            num = int(values[1])
            addr = int(values[2])
            s = self.get_sector_slice(1, num)
            r = self.rawdata[s]
            header = ObjSegment(r[0:6], 0, 0, addr, addr + 6, name="Boot Header")
            sectors = ObjSegment(r, 0, 0, addr, addr + len(r), name="Boot Sectors")
            code = ObjSegment(r[6:], 0, 0, addr + 6, addr + len(r), name="Boot Code")
            segments = [sectors, header, code]
        return segments

    def get_vtoc_segments(self):
        return []

    def get_directory_segments(self):
        return []


class AtariDiskImage(BootDiskImage):
    def __str__(self):
        return "%s Unidentified Contents" % (self.header)

    def check_size(self):
        if self.header is None:
            raise errors.InvalidDiskImage("Not a known Atari disk image format")

    def get_boot_segments(self):
        return []


def add_atr_header(bytes):
    header = AtrHeader(create=True)
    header.check_size(len(bytes))
    hlen = len(header)
    data = np.empty([hlen + len(bytes)], dtype=np.uint8)
    data[0:hlen] = header.to_array()
    data[hlen:] = bytes
    return data



# Standard library imports.
import sys
import os
import functools

# Major package imports.
import wx
import numpy as np

# Local imports.
from sawx.utils.command import Overlay
from ..utils.drawutil import get_bounds
from sawx.ui.compactgrid import MouseMode, NormalSelectMode, RectangularSelectMode

from ..commands import map as mc

import logging
log = logging.getLogger(__name__)


class EyedropperMode(RectangularSelectMode):
    icon = "eyedropper.png"
    menu_item_name = "Pick Item"
    menu_item_tooltip = "Pick an item from the grid and use as the current draw item"

    def process_left_down(self, evt):
        log.debug("EyedropperMode: process_left_down")
        cg = self.control
        input_row, input_cell = cg.main.get_row_cell_from_event(evt)
        if (input_row, input_cell) == self.last_mouse_event:
            # only process if mouse has moved to a new cell; no sub-cell
            # events!
            return
        self.last_mouse_event = (input_row, input_cell)
        row, col, _ = cg.get_row_col_from_event(evt)
        index, _ = cg.table.get_index_range(row, col)
        value = cg.segment_viewer.segment[index]
        cg.segment_viewer.set_draw_pattern(value)
        log.debug("draw_pattern=%x at index=%d %d,%d" % (value, index, row, col))
        # self.display_coords(evt, "tile=%d" % value)

    def process_mouse_motion_down(self, evt):
        self.process_left_down(evt)

    def process_left_up(self, evt):
        self.last_mouse_event = None, None


class DrawMode(RectangularSelectMode):
    icon = "shape_freehand.png"
    menu_item_name = "Draw"
    menu_item_tooltip = "Draw with current tile"

    def draw(self, evt, start=False):
        cg = self.control
        v = cg.segment_viewer
        pattern = cg.segment_viewer.draw_pattern
        print(("drawing with!", pattern, type(pattern)))
        if start:
            self.batch = mc.DrawBatchCommand()
        row, col, _ = cg.get_row_col_from_event(evt)
        if cg.main.is_inside(row, col):
            index, _ = cg.table.get_index_range(row, col)
            cmd = ChangeByteCommand(v.segment, index, index+len(pattern), pattern, False, True)
            v.editor.process_command(cmd, self.batch)

    def process_left_down(self, evt):
        self.draw(evt, True)
        self.display_coords(evt)

    def process_mouse_motion_down(self, evt):
        self.draw(evt)
        self.display_coords(evt)

    def process_left_up(self, evt):
        cg = self.control
        v = cg.segment_viewer
        v.editor.end_batch()
        self.batch = None


class OverlayMode(RectangularSelectMode):
    command = None

    def get_display_rect(self, index):
        cg = self.control
        i1 = self.start_index
        i2 = index
        if i2 < i1:
            i1, i2 = i2, i1
        (x1, y1), (x2, y2) = get_bounds(i1, i2, cg.table.items_per_row)
        extra = None
        w = x2 - x1 + 1
        h = y2 - y1 + 1
        if w > 0 or h > 0:
            extra = "rectangle: $%x x $%x" % (w, h)
        return extra

    def draw(self, evt, start=False):
        cg = self.control
        v = cg.segment_viewer
        pattern = v.draw_pattern
        row, col, _ = cg.get_row_col_from_event(evt)
        if cg.main.is_inside(row, col):
            index, _ = cg.table.get_index_range(row, col)
            if start:
                self.batch = Overlay()
                self.start_index = index
            #v.linked_base.set_caret(byte, False)
            #index = byte
            cmd = self.command(v.segment, self.start_index, index, pattern, cg.table.items_per_row)
            v.editor.process_command(cmd, self.batch)
            self.display_coords(evt, self.get_display_rect(index))

    def process_left_down(self, evt):
        self.draw(evt, True)

    def process_mouse_motion_down(self, evt):
        self.draw(evt)

    def process_left_up(self, evt):
        cg = self.control
        v = cg.segment_viewer
        v.editor.end_batch()
        self.batch = None


class LineMode(OverlayMode):
    icon = "shape_line.png"
    menu_item_name = "Line"
    menu_item_tooltip = "Draw line with current tile"
    command = mc.LineCommand


class SquareMode(OverlayMode):
    icon = "shape_hollow_square.png"
    menu_item_name = "Square"
    menu_item_tooltip = "Draw square with current tile"
    command = mc.SquareCommand


class FilledSquareMode(OverlayMode):
    icon = "shape_filled_square.png"
    menu_item_name = "Filled Square"
    menu_item_tooltip = "Draw filled square with current tile"
    command = mc.FilledSquareCommand



import os
import hashlib
import inspect
import pkg_resources

import numpy as np

from . import errors
from .utils import to_numpy, to_numpy_list, uuid

import logging
log = logging.getLogger(__name__)


class Archiver:
    archive_type = ""

    supports_multiple_containers = True

    def __str__(self):
        return self.archive_type + " archive"

    def iter_archive(self, basename, byte_data):
        yield basename, byte_data

    def pack_data(self, fh, containers, skip_missing_compressors=False):
        raise NotImplementedError


class PlainFileArchiver:
    archive_type = "plain file"

    supports_multiple_containers = False

    def __str__(self):
        return self.archive_type

    def iter_archive(self, basename, byte_data):
        yield basename, byte_data

    def pack_data(self, fh, containers, skip_missing_compressors=False):
        if len(containers) > 1:
            raise errors.InvalidArchiver(f"{str(self)} doesn't support multiple containers")
        fh.write(containers[0].calc_packed_bytes(skip_missing_compressors))


_archivers = None

def _find_archivers():
    archivers = []
    for entry_point in pkg_resources.iter_entry_points('atrip.archivers'):
        mod = entry_point.load()
        log.debug(f"find_archiver: Found module {entry_point.name}={mod.__name__}")
        for name, obj in inspect.getmembers(mod):
            if inspect.isclass(obj) and Archiver in obj.__mro__[1:]:
                log.debug(f"find_archivers:   found archiver class {name}")
                archivers.append(obj)

    return archivers

def find_archivers():
    global _archivers

    if _archivers is None:
        _archivers = _find_archivers()
    return _archivers

def find_container_items_in_archive(pathname, raw_data):
    basename = os.path.basename(pathname)
    archiver = None
    for c in find_archivers():
        items = []
        log.debug(f"trying archiver {c.archive_type}")
        try:
            archiver = c()
            items = list(archiver.iter_archive(basename, raw_data))
        except errors.InvalidArchiver as e:
            continue
        else:
            log.info(f"found archiver {c.archive_type}")
            return archiver, items
    else:
        log.info(f"image does not appear to be archived.")
    return PlainFileArchiver(), [(basename, raw_data),]



import os

import pytest
jsonpickle = pytest.importorskip("jsonpickle")
jsonpickle.set_encoder_options('json', sort_keys=True, indent=4)

import numpy as np

from atrip import Container, Segment
from atrip.container import guess_container, load


class TestJsonPickle:
    def setup(self):
        data = np.arange(4096, dtype=np.uint8)
        data[1::2] = np.repeat(np.arange(16, dtype=np.uint8), 128)
        data[::100] = 0x7f
        self.container = Container(data)
        self.container.disasm_type[100:200] = 10
        self.container.disasm_type[200:300] = 30
        self.container.disasm_type[1200:3000] = 10

    def test_simple_container(self):
        j = jsonpickle.dumps(self.container)
        print(j)
        
        c = jsonpickle.loads(j)
        j2 = jsonpickle.dumps(c)
        print(j2)

        assert j == j2

    def test_simple_segment(self):
        s = Segment(self.container)
        j = jsonpickle.dumps(s)
        print(j)
        
        c = jsonpickle.loads(j)
        j2 = jsonpickle.dumps(c)
        print(j2)

        assert j == j2

    def test_ordered_segment(self):
        order = np.arange(200, 500, dtype=np.uint32)
        order[100:200] = np.arange(0, 100, dtype=np.uint32)
        s = Segment(self.container, order)
        j = jsonpickle.dumps(s)
        print(j)
        
        c = jsonpickle.loads(j)
        j2 = jsonpickle.dumps(c)
        print(j2)

        assert j == j2

        assert np.array_equal(s.container_offset, c.container_offset)

    def test_sparse_segment(self):
        offsets = np.arange(40, dtype=np.int32) * 100
        offsets[10:20] = np.arange(70, 80, dtype=np.uint32)
        s = Segment(self.container, offsets)
        j = jsonpickle.dumps(s)
        print(j)
        
        s2 = jsonpickle.loads(j)
        j2 = jsonpickle.dumps(s2)
        print(j2)

        assert j == j2

        assert np.array_equal(s.container_offset, s2.container_offset)

    def test_file(self):
        filename = "dos_sd_test1.atr"
        pathname = os.path.join(os.path.dirname(__file__), "../samples", filename)
        container = load(pathname)
        container.guess_media_type()
        container.guess_filesystem()
        j = jsonpickle.dumps(container)
        print(j)
        
        c = jsonpickle.loads(j)
        j2 = jsonpickle.dumps(c)
        print(j2)

        assert j == j2
        print(c.segments)

    @pytest.mark.parametrize("filename", ["dos_sd_test1.atr", "dos_ed_test1.atr", "dos_dd_test1.atr", "dos33_master.dsk"])
    def test_file_with_filesytem(self, filename):
        pathname = os.path.join(os.path.dirname(__file__), "../samples", filename)
        container = load(pathname)
        container.guess_media_type()
        container.guess_filesystem()
        j = jsonpickle.dumps(container)
        print(j)
        
        c = jsonpickle.loads(j)
        j2 = jsonpickle.dumps(c)
        print(j2)

        assert j == j2
        print(c.segments)

if __name__ == "__main__":
    t = TestJsonPickle()
    t.setup()
    # t.test_simple_container()
    # t.test_simple_segment()
    # t.test_ordered_segment()
    # t.test_sparse_segment()
    t.test_file()
    # t.test_file_with_filesytem("dos33_master.dsk")
    # t.test_file_with_filesytem("dos_sd_test1.atr")



#!/usr/bin/env python
"""wxPython wrapper around texture_canvas
"""
import sys
import ctypes
import numpy as np

import wx
from wx import glcanvas

from .texture_canvas import GLSLTextureCanvas, LegacyTextureCanvas

import logging
log = logging.getLogger(__name__)
#log.setLevel(logging.DEBUG)



class wxGLSLTextureCanvas(GLSLTextureCanvas, glcanvas.GLCanvas):
    def __init__(self, parent, initial_palette, *args, **kwargs):
        """create the canvas """
        kwargs['attribList'] = (glcanvas.WX_GL_RGBA,
                                glcanvas.WX_GL_DOUBLEBUFFER,
                                glcanvas.WX_GL_MIN_ALPHA, 8, )
        glcanvas.GLCanvas.__init__(self, parent, *args, **kwargs)
        GLSLTextureCanvas.__init__(self, initial_palette)

    def ui_init_context(self):
        self.context = glcanvas.GLContext(self)

    def ui_bind_events(self):
        # execute self.onPaint whenever the parent frame is repainted
        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.Bind(wx.EVT_SIZE, self.on_size)

    def ui_set_current(self):
        self.SetCurrent(self.context)

    def ui_swap_buffers(self):
        self.SwapBuffers()

    def ui_get_window_size(self):
        w, h = self.GetClientSize()
        return w, h

    def on_paint(self, event):
        """called when window is repainted """
        # make sure we have a texture to draw
        if not self.finished_init:
            self.init_context()
        self.on_draw()


class wxLegacyTextureCanvas(LegacyTextureCanvas, wxGLSLTextureCanvas, glcanvas.GLCanvas):
    def __init__(self, *args, **kwargs):
        wxGLSLTextureCanvas.__init__(self, *args, **kwargs)


if __name__ == "__main__":
    from ..atari800.colors import NTSC

    logging.basicConfig(level=logging.DEBUG)

    def run():
        app = wx.App()
        fr = wx.Frame(None, size=(640, 480), title='wxPython OpenGL programmable pipeline demo')
        canv = wxGLSLTextureCanvas(fr, NTSC)
        fr.Show()
        app.MainLoop()
        return 0

    sys.exit(run())



import os
import sys

import wx

from ..ui.tilewrap import TileWrapControl

from ..viewer import SegmentViewer

import logging
log = logging.getLogger(__name__)


class TileWindow(wx.SplitterWindow):
    def __init__(self, parent, linked_base, *args, **kwargs):
        wx.SplitterWindow.__init__(self, parent, -1, style = wx.SP_LIVE_UPDATE)
        self._segment_viewer = None
        self.Bind(wx.EVT_SPLITTER_SASH_POS_CHANGED, self.OnSashChanged)
        self.Bind(wx.EVT_SPLITTER_SASH_POS_CHANGING, self.OnSashChanging)

        self.groups = TileWrapControl(self, linked_base, size=(256,100), command=ChangeByteCommand)

        # self.tiles = CharacterSetViewer(self, linked_base, size=(256,100))

        self.SetMinimumPaneSize(100)
        self.SetSashGravity(0.5)
        self.SplitHorizontally(self.groups, self.tiles, 0)

    @property
    def segment_viewer(self):
        return self._segment_viewer

    @segment_viewer.setter
    def segment_viewer(self, segment_viewer):
        # Need to update the child controls
        self._segment_viewer = segment_viewer
        self.groups.segment_viewer = segment_viewer
        self.tiles.segment_viewer = segment_viewer

    def OnSashChanged(self, evt):
        log.debug("sash changed to %s\n" % str(evt.GetSashPosition()))

    def OnSashChanging(self, evt):
        log.debug("sash changing to %s\n" % str(evt.GetSashPosition()))
        # uncomment this to not allow the change
        #evt.SetSashPosition(-1)

    def set_font(self):
        self.groups.set_font()
        self.tiles.set_font()

    def refresh_view(self):
        self.groups.Refresh()
        self.tiles.Refresh()

    def recalc_view(self):
        self.groups.recalc_view()
        self.tiles.recalc_view()


class TileViewer(SegmentViewer):
    name = "tile"

    ui_name = "Tile"

    has_font = True

    linked_viewer = Any(None)

    tile_groups = Any([])

    @classmethod
    def create_control(cls, parent, linked_base, mdict):
        return TileWindow(parent, linked_base, size=(160,500))

    ##### Initialization and serialization

    def from_metadata_dict(self, e):
        log.debug("metadata: %s" % str(e))
        if 'tile groups' in e:
            self.tile_groups = e['tile groups']

    def to_metadata_dict(self, mdict, document):
        mdict['tile groups'] = self.tile_groups

    ##### SegmentViewer interface

    def update_bitmap(self, evt):
        log.debug("TileViewer: machine font changed for %s" % self.control)
        if evt is not Undefined:
            self.control.set_font()
            self.control.recalc_view()

    def update_bitmap(self, evt):
        log.debug("focused viewer changed to %s" % evt)
        v = evt
        if hasattr(v, 'draw_pattern'):
            log.debug("changing to %s, checking %s" % (v, self.linked_viewer))
            if v != self.linked_viewer:
                self.linked_viewer = v
                self.control.recalc_view()
                log.debug("new linked_viewer=%s" % (self.linked_viewer))

    def show_caret(self, control, index, bit):
        # Cursor view is meaningless for this view
        pass

    def set_draw_pattern(self, value):
        # inform linked viewer that there should be a new draw pattern!
        log.debug("set draw pattern: %s in %s" % (value, self.linked_viewer))
        if self.linked_viewer is not None:
            self.linked_viewer.set_draw_pattern(value)
            self.control.groups.show_pattern(self.linked_viewer.draw_pattern)
            self.control.tiles.show_pattern(self.linked_viewer.draw_pattern)
